@@ -1,4 +1,5 @@
 {-# LANGUAGE NoMonoLocalBinds #-}
+{-# LANGUAGE FlexibleContexts #-}
 {-|
 
 Date parsing and utilities for hledger.
@@ -38,6 +39,8 @@ module Hledger.Data.Dates (
   nulldatespan,
   tests_Hledger_Data_Dates,
   failIfInvalidYear,
+  failIfInvalidMonth,
+  failIfInvalidDay,
   datesepchar,
   datesepchars,
   spanStart,
@@ -60,18 +63,20 @@ module Hledger.Data.Dates (
 )
 where
 
+import Control.Applicative ((<*))
 import Control.Monad
 import Data.List
 import Data.Maybe
 import Data.Time.Format
 import Data.Time.Calendar
 import Data.Time.Calendar.OrdinalDate
+import Data.Time.Calendar.WeekDate
 import Data.Time.Clock
 import Data.Time.LocalTime
 import Safe (headMay, lastMay, readMay)
 import System.Locale (defaultTimeLocale)
 import Test.HUnit
-import Text.ParserCombinators.Parsec
+import Text.Parsec
 import Text.Printf
 
 import Hledger.Data.Types
@@ -85,7 +90,58 @@ instance Show DateSpan where
 showDate :: Day -> String
 showDate = formatTime defaultTimeLocale "%0C%y/%m/%d"
 
-showDateSpan (DateSpan from to) =
+-- XXX review for more boundary crossing issues
+-- | Render a datespan as a display string, abbreviating into a
+-- compact form if possible.
+showDateSpan ds@(DateSpan (Just from) (Just to)) =
+  case (toGregorian from, toGregorian to) of
+    -- special cases we can abbreviate:
+    -- a year, YYYY
+    ((fy,1,1), (ty,1,1))   | fy+1==ty           -> formatTime defaultTimeLocale "%0C%y" from
+    -- a half, YYYYhN
+    ((fy,1,1), (ty,7,1))   | fy==ty             -> formatTime defaultTimeLocale "%0C%yh1" from
+    ((fy,7,1), (ty,1,1))   | fy+1==ty           -> formatTime defaultTimeLocale "%0C%yh2" from
+    -- a quarter, YYYYqN
+    ((fy,1,1), (ty,4,1))   | fy==ty             -> formatTime defaultTimeLocale "%0C%yq1" from
+    ((fy,4,1), (ty,7,1))   | fy==ty             -> formatTime defaultTimeLocale "%0C%yq2" from
+    ((fy,7,1), (ty,10,1))  | fy==ty             -> formatTime defaultTimeLocale "%0C%yq3" from
+    ((fy,10,1), (ty,1,1))  | fy+1==ty           -> formatTime defaultTimeLocale "%0C%yq4" from
+    -- a month, YYYY/MM
+    ((fy,fm,1), (ty,tm,1)) | fy==ty && fm+1==tm -> formatTime defaultTimeLocale "%0C%y/%m" from
+    ((fy,12,1), (ty,1,1))  | fy+1==ty           -> formatTime defaultTimeLocale "%0C%y/%m" from
+    -- a week (two successive mondays),
+    -- YYYYwN ("week N of year YYYY")
+    -- _ | let ((fy,fw,fd), (ty,tw,td)) = (toWeekDate from, toWeekDate to) in fy==ty && fw+1==tw && fd==1 && td==1
+    --                                             -> formatTime defaultTimeLocale "%0f%gw%V" from
+    -- YYYY/MM/DDwN ("week N, starting on YYYY/MM/DD")
+    _ | let ((fy,fw,fd), (ty,tw,td)) = (toWeekDate from, toWeekDate (addDays (-1) to)) in fy==ty && fw==tw && fd==1 && td==7
+                                                -> formatTime defaultTimeLocale "%0C%y/%m/%dw%V" from
+    -- a day, YYYY/MM/DDd (d suffix is to distinguish from a regular date in register)
+    ((fy,fm,fd), (ty,tm,td)) | fy==ty && fm==tm && fd+1==td -> formatTime defaultTimeLocale "%0C%y/%m/%dd" from
+    -- crossing a year boundary
+    ((fy,fm,fd), (ty,tm,td)) | fy+1==ty && fm==12 && tm==1 && fd==31 && td==1 -> formatTime defaultTimeLocale "%0C%y/%m/%dd" from
+    -- crossing a month boundary XXX wrongly shows LEAPYEAR/2/28-LEAPYEAR/3/1 as LEAPYEAR/2/28
+    ((fy,fm,fd), (ty,tm,td)) | fy==ty && fm+1==tm && fd `elem` fromMaybe [] (lookup fm lastdayofmonth) && td==1 -> formatTime defaultTimeLocale "%0C%y/%m/%dd" from
+    -- otherwise, YYYY/MM/DD-YYYY/MM/DD
+    _                                           -> showDateSpan' ds
+  where lastdayofmonth = [(1,[31])
+                         ,(2,[28,29])
+                         ,(3,[31])
+                         ,(4,[30])
+                         ,(5,[30])
+                         ,(6,[31])
+                         ,(7,[31])
+                         ,(8,[31])
+                         ,(9,[30])
+                         ,(10,[31])
+                         ,(11,[30])
+                         ,(12,[31])
+                         ]
+
+showDateSpan ds = showDateSpan' ds
+
+-- | Render a datespan as a display string.
+showDateSpan' (DateSpan from to) =
   concat
     [maybe "" showDate from
     ,"-"
@@ -119,7 +175,7 @@ spanStart (DateSpan d _) = d
 spanEnd :: DateSpan -> Maybe Day
 spanEnd (DateSpan _ d) = d
 
--- might be useful later: http://en.wikipedia.org/wiki/Allen%27s_interval_algebra 
+-- might be useful later: http://en.wikipedia.org/wiki/Allen%27s_interval_algebra
 
 -- | Get overall span enclosing multiple sequentially ordered spans.
 spansSpan :: [DateSpan] -> DateSpan
@@ -170,7 +226,7 @@ spanContainsDate (DateSpan Nothing Nothing)   _ = True
 spanContainsDate (DateSpan Nothing (Just e))  d = d < e
 spanContainsDate (DateSpan (Just b) Nothing)  d = d >= b
 spanContainsDate (DateSpan (Just b) (Just e)) d = d >= b && d < e
-    
+
 -- | Calculate the intersection of a number of datespans.
 spansIntersect [] = nulldatespan
 spansIntersect [d] = d
@@ -210,7 +266,7 @@ earliest (Just d1) (Just d2) = Just $ min d1 d2
 -- | Parse a period expression to an Interval and overall DateSpan using
 -- the provided reference date, or return a parse error.
 parsePeriodExpr :: Day -> String -> Either ParseError (Interval, DateSpan)
-parsePeriodExpr refdate = parsewith (periodexpr refdate)
+parsePeriodExpr refdate = parsewith (periodexpr refdate <* eof)
 
 maybePeriod :: Day -> String -> Maybe (Interval,DateSpan)
 maybePeriod refdate = either (const Nothing) Just . parsePeriodExpr refdate
@@ -221,7 +277,7 @@ maybePeriod refdate = either (const Nothing) Just . parsePeriodExpr refdate
 -- dateSpanAsText (DateSpan Nothing (Just e))  = printf "to %s" (show e)
 -- dateSpanAsText (DateSpan (Just b) Nothing)  = printf "from %s" (show b)
 -- dateSpanAsText (DateSpan (Just b) (Just e)) = printf "%s to %s" (show b) (show e)
-    
+
 -- | Convert a single smart date string to a date span using the provided
 -- reference date, or raise an error.
 -- spanFromSmartDateString :: Day -> String -> DateSpan
@@ -366,9 +422,9 @@ nthdayofweekcontaining n d | d1 >= d    = d1
 
 -- | Parse a couple of date string formats to a time type.
 parsedateM :: String -> Maybe Day
-parsedateM s = firstJust [ 
+parsedateM s = firstJust [
      parseTime defaultTimeLocale "%Y/%m/%d" s,
-     parseTime defaultTimeLocale "%Y-%m-%d" s 
+     parseTime defaultTimeLocale "%Y-%m-%d" s
      ]
 
 -- -- | Parse a date-time string to a time type, or raise an error.
@@ -386,7 +442,7 @@ parsedate s =  fromMaybe (error' $ "could not parse date \"" ++ s ++ "\"")
 parsetimewith :: ParseTime t => String -> String -> t -> t
 parsetimewith pat s def = fromMaybe def $ parseTime defaultTimeLocale pat s
 
-{-| 
+{-|
 Parse a date in any of the formats allowed in ledger's period expressions,
 and maybe some others:
 
@@ -402,14 +458,14 @@ and maybe some others:
 Returns a SmartDate, to be converted to a full date later (see fixSmartDate).
 Assumes any text in the parse stream has been lowercased.
 -}
-smartdate :: GenParser Char st SmartDate
+smartdate :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 smartdate = do
   -- XXX maybe obscures date errors ? see ledgerdate
   (y,m,d) <- choice' [yyyymmdd, ymd, ym, md, y, d, month, mon, today, yesterday, tomorrow, lastthisnextthing]
   return (y,m,d)
 
 -- | Like smartdate, but there must be nothing other than whitespace after the date.
-smartdateonly :: GenParser Char st SmartDate
+smartdateonly :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 smartdateonly = do
   d <- smartdate
   many spacenonewline
@@ -417,6 +473,7 @@ smartdateonly = do
   return d
 
 datesepchars = "/-."
+datesepchar :: Stream [Char] m Char => ParsecT [Char] st m Char
 datesepchar = oneOf datesepchars
 
 validYear, validMonth, validDay :: String -> Bool
@@ -429,7 +486,7 @@ failIfInvalidYear s  = unless (validYear s)  $ fail $ "bad year number: " ++ s
 failIfInvalidMonth s = unless (validMonth s) $ fail $ "bad month number: " ++ s
 failIfInvalidDay s   = unless (validDay s)   $ fail $ "bad day number: " ++ s
 
-yyyymmdd :: GenParser Char st SmartDate
+yyyymmdd :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 yyyymmdd = do
   y <- count 4 digit
   m <- count 2 digit
@@ -438,19 +495,19 @@ yyyymmdd = do
   failIfInvalidDay d
   return (y,m,d)
 
-ymd :: GenParser Char st SmartDate
+ymd :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 ymd = do
   y <- many1 digit
   failIfInvalidYear y
-  datesepchar
+  sep <- datesepchar
   m <- many1 digit
   failIfInvalidMonth m
-  datesepchar
+  char sep
   d <- many1 digit
   failIfInvalidDay d
   return $ (y,m,d)
 
-ym :: GenParser Char st SmartDate
+ym :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 ym = do
   y <- many1 digit
   failIfInvalidYear y
@@ -459,19 +516,19 @@ ym = do
   failIfInvalidMonth m
   return (y,m,"")
 
-y :: GenParser Char st SmartDate
+y :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 y = do
   y <- many1 digit
   failIfInvalidYear y
   return (y,"","")
 
-d :: GenParser Char st SmartDate
+d :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 d = do
   d <- many1 digit
   failIfInvalidDay d
   return ("","",d)
 
-md :: GenParser Char st SmartDate
+md :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 md = do
   m <- many1 digit
   failIfInvalidMonth m
@@ -489,24 +546,24 @@ monthabbrevs   = ["jan","feb","mar","apr","may","jun","jul","aug","sep","oct","n
 monthIndex s = maybe 0 (+1) $ lowercase s `elemIndex` months
 monIndex s   = maybe 0 (+1) $ lowercase s `elemIndex` monthabbrevs
 
-month :: GenParser Char st SmartDate
+month :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 month = do
   m <- choice $ map (try . string) months
   let i = monthIndex m
   return ("",show i,"")
 
-mon :: GenParser Char st SmartDate
+mon :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 mon = do
   m <- choice $ map (try . string) monthabbrevs
   let i = monIndex m
   return ("",show i,"")
 
-today,yesterday,tomorrow :: GenParser Char st SmartDate
+today,yesterday,tomorrow :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 today     = string "today"     >> return ("","","today")
 yesterday = string "yesterday" >> return ("","","yesterday")
 tomorrow  = string "tomorrow"  >> return ("","","tomorrow")
 
-lastthisnextthing :: GenParser Char st SmartDate
+lastthisnextthing :: Stream [Char] m Char => ParsecT [Char] st m SmartDate
 lastthisnextthing = do
   r <- choice [
         string "last"
@@ -523,10 +580,10 @@ lastthisnextthing = do
       ]
 -- XXX support these in fixSmartDate
 --       ++ (map string $ months ++ monthabbrevs ++ weekdays ++ weekdayabbrevs)
-            
+
   return ("",r,p)
 
-periodexpr :: Day -> GenParser Char st (Interval, DateSpan)
+periodexpr :: Stream [Char] m Char => Day -> ParsecT [Char] st m (Interval, DateSpan)
 periodexpr rdate = choice $ map try [
                     intervalanddateperiodexpr rdate,
                     intervalperiodexpr,
@@ -534,7 +591,7 @@ periodexpr rdate = choice $ map try [
                     (return (NoInterval,DateSpan Nothing Nothing))
                    ]
 
-intervalanddateperiodexpr :: Day -> GenParser Char st (Interval, DateSpan)
+intervalanddateperiodexpr :: Stream [Char] m Char => Day -> ParsecT [Char] st m (Interval, DateSpan)
 intervalanddateperiodexpr rdate = do
   many spacenonewline
   i <- reportinginterval
@@ -542,20 +599,20 @@ intervalanddateperiodexpr rdate = do
   s <- periodexprdatespan rdate
   return (i,s)
 
-intervalperiodexpr :: GenParser Char st (Interval, DateSpan)
+intervalperiodexpr :: Stream [Char] m Char => ParsecT [Char] st m (Interval, DateSpan)
 intervalperiodexpr = do
   many spacenonewline
   i <- reportinginterval
   return (i, DateSpan Nothing Nothing)
 
-dateperiodexpr :: Day -> GenParser Char st (Interval, DateSpan)
+dateperiodexpr :: Stream [Char] m Char => Day -> ParsecT [Char] st m (Interval, DateSpan)
 dateperiodexpr rdate = do
   many spacenonewline
   s <- periodexprdatespan rdate
   return (NoInterval, s)
 
 -- Parse a reporting interval.
-reportinginterval :: GenParser Char st Interval
+reportinginterval :: Stream [Char] m Char => ParsecT [Char] st m Interval
 reportinginterval = choice' [
                        tryinterval "day"     "daily"     Days,
                        tryinterval "week"    "weekly"    Weeks,
@@ -595,7 +652,7 @@ reportinginterval = choice' [
       thsuffix = choice' $ map string ["st","nd","rd","th"]
 
       -- Parse any of several variants of a basic interval, eg "daily", "every day", "every N days".
-      tryinterval :: String -> String -> (Int -> Interval) -> GenParser Char st Interval
+      tryinterval :: Stream [Char] m Char => String -> String -> (Int -> Interval) -> ParsecT [Char] st m Interval
       tryinterval singular compact intcons =
           choice' [
            do string compact
@@ -613,7 +670,7 @@ reportinginterval = choice' [
            ]
           where plural = singular ++ "s"
 
-periodexprdatespan :: Day -> GenParser Char st DateSpan
+periodexprdatespan :: Stream [Char] m Char => Day -> ParsecT [Char] st m DateSpan
 periodexprdatespan rdate = choice $ map try [
                             doubledatespan rdate,
                             fromdatespan rdate,
@@ -621,7 +678,7 @@ periodexprdatespan rdate = choice $ map try [
                             justdatespan rdate
                            ]
 
-doubledatespan :: Day -> GenParser Char st DateSpan
+doubledatespan :: Stream [Char] m Char => Day -> ParsecT [Char] st m DateSpan
 doubledatespan rdate = do
   optional (string "from" >> many spacenonewline)
   b <- smartdate
@@ -630,7 +687,7 @@ doubledatespan rdate = do
   e <- smartdate
   return $ DateSpan (Just $ fixSmartDate rdate b) (Just $ fixSmartDate rdate e)
 
-fromdatespan :: Day -> GenParser Char st DateSpan
+fromdatespan :: Stream [Char] m Char => Day -> ParsecT [Char] st m DateSpan
 fromdatespan rdate = do
   b <- choice [
     do
@@ -644,13 +701,13 @@ fromdatespan rdate = do
     ]
   return $ DateSpan (Just $ fixSmartDate rdate b) Nothing
 
-todatespan :: Day -> GenParser Char st DateSpan
+todatespan :: Stream [Char] m Char => Day -> ParsecT [Char] st m DateSpan
 todatespan rdate = do
   choice [string "to", string "-"] >> many spacenonewline
   e <- smartdate
   return $ DateSpan Nothing (Just $ fixSmartDate rdate e)
 
-justdatespan :: Day -> GenParser Char st DateSpan
+justdatespan :: Stream [Char] m Char => Day -> ParsecT [Char] st m DateSpan
 justdatespan rdate = do
   optional (string "in" >> many spacenonewline)
   d <- smartdate
