@@ -22,7 +22,8 @@ module Hledger.Data.Journal (
   -- * Filtering
   filterJournalTransactions,
   filterJournalPostings,
-  filterJournalPostingAmounts,
+  filterJournalAmounts,
+  filterTransactionAmounts,
   filterPostingAmount,
   -- * Querying
   journalAccountNames,
@@ -45,6 +46,7 @@ module Hledger.Data.Journal (
   journalEquityAccountQuery,
   journalCashAccountQuery,
   -- * Misc
+  canonicalStyles,
   matchpats,
   nullctx,
   nulljournal,
@@ -58,6 +60,7 @@ import Data.List
 -- import Data.Map (findWithDefault)
 import Data.Maybe
 import Data.Ord
+import Safe (headMay)
 import Data.Time.Calendar
 import Data.Time.LocalTime
 import Data.Tree
@@ -131,7 +134,7 @@ nulljournal = Journal { jmodifiertxns = []
                       }
 
 nullctx :: JournalContext
-nullctx = Ctx { ctxYear = Nothing, ctxCommodityAndStyle = Nothing, ctxAccount = [], ctxAliases = [] }
+nullctx = Ctx { ctxYear = Nothing, ctxDefaultCommodityAndStyle = Nothing, ctxAccount = [], ctxAliases = [] }
 
 journalFilePath :: Journal -> FilePath
 journalFilePath = fst . mainfile
@@ -143,19 +146,19 @@ mainfile :: Journal -> (FilePath, String)
 mainfile = headDef ("", "") . files
 
 addTransaction :: Transaction -> Journal -> Journal
-addTransaction t l0 = l0 { jtxns = t : jtxns l0 }
+addTransaction t j = j { jtxns = t : jtxns j }
 
 addModifierTransaction :: ModifierTransaction -> Journal -> Journal
-addModifierTransaction mt l0 = l0 { jmodifiertxns = mt : jmodifiertxns l0 }
+addModifierTransaction mt j = j { jmodifiertxns = mt : jmodifiertxns j }
 
 addPeriodicTransaction :: PeriodicTransaction -> Journal -> Journal
-addPeriodicTransaction pt l0 = l0 { jperiodictxns = pt : jperiodictxns l0 }
+addPeriodicTransaction pt j = j { jperiodictxns = pt : jperiodictxns j }
 
 addHistoricalPrice :: HistoricalPrice -> Journal -> Journal
-addHistoricalPrice h l0 = l0 { historical_prices = h : historical_prices l0 }
+addHistoricalPrice h j = j { historical_prices = h : historical_prices j }
 
 addTimeLogEntry :: TimeLogEntry -> Journal -> Journal
-addTimeLogEntry tle l0 = l0 { open_timelog_entries = tle : open_timelog_entries l0 }
+addTimeLogEntry tle j = j { open_timelog_entries = tle : open_timelog_entries j }
 
 -- | Unique transaction descriptions used in this journal.
 journalDescriptions :: Journal -> [String]
@@ -231,6 +234,10 @@ journalCashAccountQuery j = And [journalAssetAccountQuery j, Not $ Acct "(receiv
 -------------------------------------------------------------------------------
 -- filtering V2
 
+-- | Keep only transactions matching the query expression.
+filterJournalTransactions :: Query -> Journal -> Journal
+filterJournalTransactions q j@Journal{jtxns=ts} = j{jtxns=filter (q `matchesTransaction`) ts}
+
 -- | Keep only postings matching the query expression.
 -- This can leave unbalanced transactions.
 filterJournalPostings :: Query -> Journal -> Journal
@@ -238,21 +245,20 @@ filterJournalPostings q j@Journal{jtxns=ts} = j{jtxns=map filtertransactionposti
     where
       filtertransactionpostings t@Transaction{tpostings=ps} = t{tpostings=filter (q `matchesPosting`) ps}
 
--- Within each posting's amount, keep only the parts matching the query.
+-- | Within each posting's amount, keep only the parts matching the query.
 -- This can leave unbalanced transactions.
-filterJournalPostingAmounts :: Query -> Journal -> Journal
-filterJournalPostingAmounts q j@Journal{jtxns=ts} = j{jtxns=map filtertransactionpostings ts}
-    where
-      filtertransactionpostings t@Transaction{tpostings=ps} = t{tpostings=map (filterPostingAmount q) ps}
+filterJournalAmounts :: Query -> Journal -> Journal
+filterJournalAmounts q j@Journal{jtxns=ts} = j{jtxns=map (filterTransactionAmounts q) ts}
+
+-- | Filter out all parts of this transaction's amounts which do not match the query.
+-- This can leave the transaction unbalanced.
+filterTransactionAmounts :: Query -> Transaction -> Transaction
+filterTransactionAmounts q t@Transaction{tpostings=ps} = t{tpostings=map (filterPostingAmount q) ps}
 
 -- | Filter out all parts of this posting's amount which do not match the query.
 filterPostingAmount :: Query -> Posting -> Posting
 filterPostingAmount q p@Posting{pamount=Mixed as} = p{pamount=Mixed $ filter (q `matchesAmount`) as}
 
--- | Keep only transactions matching the query expression.
-filterJournalTransactions :: Query -> Journal -> Journal
-filterJournalTransactions q j@Journal{jtxns=ts} = j{jtxns=filter (q `matchesTransaction`) ts}
-
 {-
 -------------------------------------------------------------------------------
 -- filtering V1
@@ -379,22 +385,36 @@ filterJournalTransactionsByAccount apats j@Journal{jtxns=ts} = j{jtxns=filter tm
 -}
 
 -- | Apply additional account aliases (eg from the command-line) to all postings in a journal.
-journalApplyAliases :: [(AccountName,AccountName)] -> Journal -> Journal
-journalApplyAliases aliases j@Journal{jtxns=ts} = j{jtxns=map fixtransaction ts}
+journalApplyAliases :: [AccountAlias] -> Journal -> Journal
+journalApplyAliases aliases j@Journal{jtxns=ts} =
+  -- (if null aliases
+  --  then id
+  --  else (dbgtrace $
+  --        "applying additional command-line aliases:\n"
+  --        ++ chomp (unlines $ map (" "++) $ lines $ ppShow aliases))) $
+  j{jtxns=map fixtransaction ts}
     where
       fixtransaction t@Transaction{tpostings=ps} = t{tpostings=map fixposting ps}
       fixposting p@Posting{paccount=a} = p{paccount=accountNameApplyAliases aliases a}
 
 -- | Do post-parse processing on a journal to make it ready for use: check
 -- all transactions balance, canonicalise amount formats, close any open
--- timelog entries and so on.
-journalFinalise :: ClockTime -> LocalTime -> FilePath -> String -> JournalContext -> Journal -> Either String Journal
-journalFinalise tclock tlocal path txt ctx j@Journal{files=fs} = do
+-- timelog entries, maybe check balance assertions and so on.
+journalFinalise :: ClockTime -> LocalTime -> FilePath -> String -> JournalContext -> Bool -> Journal -> Either String Journal
+journalFinalise tclock tlocal path txt ctx assrt j@Journal{files=fs} = do
   (journalBalanceTransactions $
     journalCanonicaliseAmounts $
-    journalCloseTimeLogEntries tlocal
-    j{files=(path,txt):fs, filereadtime=tclock, jContext=ctx})
-  >>= journalCheckBalanceAssertions
+    journalCloseTimeLogEntries tlocal $
+    j{ files=(path,txt):fs
+     , filereadtime=tclock
+     , jContext=ctx
+     , jtxns=reverse $ jtxns j -- NOTE: see addTransaction
+     , jmodifiertxns=reverse $ jmodifiertxns j -- NOTE: see addModifierTransaction
+     , jperiodictxns=reverse $ jperiodictxns j -- NOTE: see addPeriodicTransaction
+     , historical_prices=reverse $ historical_prices j -- NOTE: see addHistoricalPrice
+     , open_timelog_entries=reverse $ open_timelog_entries j -- NOTE: see addTimeLogEntry
+     })
+  >>= if assrt then journalCheckBalanceAssertions else return
 
 -- | Check any balance assertions in the journal and return an error
 -- message if any of them fail.
@@ -423,26 +443,28 @@ checkBalanceAssertionsForAccount ps
 -- If it does, return the new balance, otherwise add an error to the
 -- error list. Intended to be called from a fold.
 checkBalanceAssertion :: ([String],MixedAmount) -> [Posting] -> ([String],MixedAmount)
-checkBalanceAssertion (errs,bal) ps
-  | null ps = (errs,bal)
-  | isNothing assertion = (errs,bal)
-  |
-    -- bal' /= assertedbal  -- MixedAmount's Eq instance currently gets confused by different precisions
-    not $ isReallyZeroMixedAmount (bal' - assertedbal)
-      = (errs++[err], bal')
-  | otherwise = (errs,bal')
+checkBalanceAssertion (errs,startbal) ps
+  | null ps = (errs,startbal)
+  | isNothing assertion = (errs,startbal)
+  | -- bal' /= assertedbal  -- MixedAmount's Eq instance currently gets confused by different precisions
+    not $ isReallyZeroMixedAmount (bal - assertedbal) = (errs++[err], fullbal)
+  | otherwise = (errs,fullbal)
   where
     p = last ps
     assertion = pbalanceassertion p
-    Just assertedbal = assertion
-    bal' = sum $ [bal] ++ map pamount ps
-    err = printf "Balance assertion failed for account %s on %s\n%sAfter posting:\n   %s\nexpected balance is %s, actual balance was %s."
+    Just assertedbal = dbg2 "assertedbal" assertion
+    assertedcomm = dbg2 "assertedcomm" $ maybe "" acommodity $ headMay $ amounts assertedbal
+    fullbal = dbg2 "fullbal" $ sum $ [dbg2 "startbal" startbal] ++ map pamount ps
+    singlebal = dbg2 "singlebal" $ filterMixedAmount (\a -> acommodity a == assertedcomm) fullbal
+    bal = singlebal -- check single-commodity balance like Ledger; maybe add == FULLBAL later
+    err = printf "Balance assertion failed for account %s on %s\n%sAfter posting:\n   %s\nexpected balance in commodity \"%s\" is %s, calculated balance was %s."
                  (paccount p)
                  (show $ postingDate p)
                  (maybe "" (("In transaction:\n"++).show) $ ptransaction p)
                  (show p)
+                 assertedcomm
                  (showMixedAmount assertedbal)
-                 (showMixedAmount bal')
+                 (showMixedAmount singlebal)
 
 -- Given a sequence of postings to a single account, split it into
 -- sub-sequences consisting of ordinary postings followed by a single
@@ -450,11 +472,11 @@ checkBalanceAssertion (errs,bal) ps
 -- assertion are discarded.
 splitAssertions :: [Posting] -> [[Posting]]
 splitAssertions ps
-  | null rest = [[]]
+  | null rest = []
   | otherwise = (ps'++[head rest]):splitAssertions (tail rest)
   where
     (ps',rest) = break (isJust . pbalanceassertion) ps
-    
+
 -- | Fill in any missing amounts and check that all journal transactions
 -- balance, or return an error message. This is done after parsing all
 -- amounts and working out the canonical commodities, since balancing
@@ -473,12 +495,34 @@ journalCanonicaliseAmounts :: Journal -> Journal
 journalCanonicaliseAmounts j@Journal{jtxns=ts} = j''
     where
       j'' = j'{jtxns=map fixtransaction ts}
-      j' = j{jcommoditystyles = canonicalStyles $ journalAmounts j}
+      j' = j{jcommoditystyles = canonicalStyles $ dbgAt 8 "journalAmounts" $ journalAmounts j}
       fixtransaction t@Transaction{tpostings=ps} = t{tpostings=map fixposting ps}
       fixposting p@Posting{pamount=a} = p{pamount=fixmixedamount a}
       fixmixedamount (Mixed as) = Mixed $ map fixamount as
       fixamount a@Amount{acommodity=c} = a{astyle=journalCommodityStyle j' c}
 
+-- | Given a list of amounts in parse order, build a map from commodities
+-- to canonical display styles for amounts in that commodity.
+canonicalStyles :: [Amount] -> M.Map Commodity AmountStyle
+canonicalStyles amts = M.fromList commstyles
+  where
+    samecomm = \a1 a2 -> acommodity a1 == acommodity a2
+    commamts = [(acommodity $ head as, as) | as <- groupBy samecomm $ sortBy (comparing acommodity) amts]
+    commstyles = [(c, canonicalStyleFrom $ map astyle as) | (c,as) <- commamts]
+
+-- Given an ordered list of amount styles for a commodity, build a canonical style.
+canonicalStyleFrom :: [AmountStyle] -> AmountStyle
+canonicalStyleFrom [] = amountstyle
+canonicalStyleFrom ss@(first:_) =
+  first{asprecision=prec, asdecimalpoint=mdec, asdigitgroups=mgrps}
+  where
+    -- precision is the maximum of all precisions seen
+    prec = maximum $ map asprecision ss
+    -- find the first decimal point and the first digit group style seen,
+    -- or use defaults.
+    mdec  = Just $ headDef '.' $ catMaybes $ map asdecimalpoint ss
+    mgrps = maybe Nothing Just $ headMay $ catMaybes $ map asdigitgroups ss
+
 -- | Get this journal's canonical amount style for the given commodity, or the null style.
 journalCommodityStyle :: Journal -> Commodity -> AmountStyle
 journalCommodityStyle j c = M.findWithDefault amountstyle c $ jcommoditystyles j
@@ -623,10 +667,11 @@ abspat pat = if isnegativepat pat then drop (length negateprefix) pat else pat
 --     liabilities:debts  $1
 --     assets:bank:checking
 --
-Right samplejournal = journalBalanceTransactions $ 
+Right samplejournal = journalBalanceTransactions $
          nulljournal
          {jtxns = [
            txnTieKnot $ Transaction {
+             tsourcepos=nullsourcepos,
              tdate=parsedate "2008/01/01",
              tdate2=Nothing,
              tstatus=False,
@@ -642,6 +687,7 @@ Right samplejournal = journalBalanceTransactions $
            }
           ,
            txnTieKnot $ Transaction {
+             tsourcepos=nullsourcepos,
              tdate=parsedate "2008/06/01",
              tdate2=Nothing,
              tstatus=False,
@@ -657,6 +703,7 @@ Right samplejournal = journalBalanceTransactions $
            }
           ,
            txnTieKnot $ Transaction {
+             tsourcepos=nullsourcepos,
              tdate=parsedate "2008/06/02",
              tdate2=Nothing,
              tstatus=False,
@@ -672,6 +719,7 @@ Right samplejournal = journalBalanceTransactions $
            }
           ,
            txnTieKnot $ Transaction {
+             tsourcepos=nullsourcepos,
              tdate=parsedate "2008/06/03",
              tdate2=Nothing,
              tstatus=True,
@@ -687,6 +735,7 @@ Right samplejournal = journalBalanceTransactions $
            }
           ,
            txnTieKnot $ Transaction {
+             tsourcepos=nullsourcepos,
              tdate=parsedate "2008/12/31",
              tdate2=Nothing,
              tstatus=False,
