@@ -37,6 +37,7 @@ module Hledger.Data.Posting (
   joinAccountNames,
   concatAccountNames,
   accountNameApplyAliases,
+  accountNameApplyOneAlias,
   -- * arithmetic
   sumPostings,
   -- * rendering
@@ -79,11 +80,12 @@ nullposting = Posting
 posting = nullposting
 
 post :: AccountName -> Amount -> Posting
-post acct amt = posting {paccount=acct, pamount=mixed amt}
+post acct amt = posting {paccount=acct, pamount=Mixed [amt]}
 
+-- XXX once rendered user output, but just for debugging now; clean up
 showPosting :: Posting -> String
 showPosting p@Posting{paccount=a,pamount=amt,ptype=t} =
-    unlines $ [concatTopPadded [showaccountname a ++ " ", showamount amt, showComment (pcomment p)]]
+    unlines $ [concatTopPadded [show (postingDate p) ++ " ", showaccountname a ++ " ", showamount amt, showComment (pcomment p)]]
     where
       ledger3ishlayout = False
       acctnamewidth = if ledger3ishlayout then 25 else 22
@@ -121,7 +123,7 @@ sumPostings = sum . map pamount
 -- there is no parent transaction.
 postingDate :: Posting -> Day
 postingDate p = fromMaybe txndate $ pdate p
-    where 
+    where
       txndate = maybe nulldate tdate $ ptransaction p
 
 -- | Get a posting's secondary (secondary) date, which is the first of:
@@ -217,14 +219,22 @@ concatAccountNames :: [AccountName] -> AccountName
 concatAccountNames as = accountNameWithPostingType t $ intercalate ":" $ map accountNameWithoutPostingType as
     where t = headDef RegularPosting $ filter (/= RegularPosting) $ map accountNamePostingType as
 
+-- | Rewrite an account name using all applicable aliases from the given list, in sequence.
+accountNameApplyAliases :: [AccountAlias] -> AccountName -> AccountName
+accountNameApplyAliases aliases a = accountNameWithPostingType atype aname'
+  where
+    (aname,atype) = (accountNameWithoutPostingType a, accountNamePostingType a)
+    matchingaliases = filter (\(re,_) -> regexMatchesCI re aname) aliases
+    aname' = foldl (flip (uncurry regexReplaceCI)) aname matchingaliases
+
 -- | Rewrite an account name using the first applicable alias from the given list, if any.
-accountNameApplyAliases :: [(AccountName,AccountName)] -> AccountName -> AccountName
-accountNameApplyAliases aliases a = withorigtype
-    where
-      (a',t) = (accountNameWithoutPostingType a, accountNamePostingType a)
-      firstmatchingalias = headDef Nothing $ map Just $ filter (\(orig,_) -> orig == a' || orig `isAccountNamePrefixOf` a') aliases
-      rewritten = maybe a' (\(orig,alias) -> alias++drop (length orig) a') firstmatchingalias
-      withorigtype = accountNameWithPostingType t rewritten
+accountNameApplyOneAlias :: [AccountAlias] -> AccountName -> AccountName
+accountNameApplyOneAlias aliases a = accountNameWithPostingType atype aname'
+  where
+    (aname,atype) = (accountNameWithoutPostingType a, accountNamePostingType a)
+    firstmatchingalias = headDef Nothing $ map Just $ filter (\(re,_) -> regexMatchesCI re aname) aliases
+    applyAlias = uncurry regexReplaceCI
+    aname' = maybe id applyAlias firstmatchingalias $ aname
 
 tests_Hledger_Data_Posting = TestList [
 
@@ -250,4 +260,4 @@ tests_Hledger_Data_Posting = TestList [
     concatAccountNames ["a","(b)","[c:d]"] `is` "(a:b:c:d)"
 
  ]
- 
+
