@@ -24,17 +24,17 @@ import Hledger.Data.Amount
 import Hledger.Data.Posting
 import Hledger.Data.Transaction
 
-instance Show TimeLogEntry where 
+instance Show TimeLogEntry where
     show t = printf "%s %s %s" (show $ tlcode t) (show $ tldatetime t) (tlcomment t)
 
-instance Show TimeLogCode where 
+instance Show TimeLogCode where
     show SetBalance = "b"
     show SetRequiredHours = "h"
     show In = "i"
     show Out = "o"
     show FinalOut = "O"
 
-instance Read TimeLogCode where 
+instance Read TimeLogCode where
     readsPrec _ ('b' : xs) = [(SetBalance, xs)]
     readsPrec _ ('h' : xs) = [(SetRequiredHours, xs)]
     readsPrec _ ('i' : xs) = [(In, xs)]
@@ -51,7 +51,7 @@ timeLogEntriesToTransactions now [i]
     | odate > idate = entryFromTimeLogInOut i o' : timeLogEntriesToTransactions now [i',o]
     | otherwise = [entryFromTimeLogInOut i o]
     where
-      o = TimeLogEntry Out end ""
+      o = TimeLogEntry (tlsourcepos i) Out end ""
       end = if itime > now then itime else now
       (itime,otime) = (tldatetime i,tldatetime o)
       (idate,odate) = (localDay itime,localDay otime)
@@ -72,18 +72,19 @@ timeLogEntriesToTransactions now (i:o:rest)
 entryFromTimeLogInOut :: TimeLogEntry -> TimeLogEntry -> Transaction
 entryFromTimeLogInOut i o
     | otime >= itime = t
-    | otherwise = 
+    | otherwise =
         error' $ "clock-out time less than clock-in time in:\n" ++ showTransaction t
     where
       t = Transaction {
-            tdate         = idate,
-            tdate2 = Nothing,
-            tstatus       = True,
-            tcode         = "",
-            tdescription  = showtime itod ++ "-" ++ showtime otod,
-            tcomment      = "",
-            ttags     = [],
-            tpostings = ps,
+            tsourcepos   = tlsourcepos i,
+            tdate        = idate,
+            tdate2       = Nothing,
+            tstatus      = True,
+            tcode        = "",
+            tdescription = showtime itod ++ "-" ++ showtime otod,
+            tcomment     = "",
+            ttags        = [],
+            tpostings    = ps,
             tpreceding_comment_lines=""
           }
       showtime = take 5 . show
@@ -107,7 +108,7 @@ tests_Hledger_Data_TimeLog = TestList [
      let now = utcToLocalTime tz now'
          nowstr = showtime now
          yesterday = prevday today
-         clockin = TimeLogEntry In
+         clockin = TimeLogEntry nullsourcepos In
          mktime d = LocalTime d . fromMaybe midnight . parseTime defaultTimeLocale "%H:%M:%S"
          showtime = formatTime defaultTimeLocale "%H:%M"
          assertEntriesGiveStrings name es ss = assertEqual name ss (map tdescription $ timeLogEntriesToTransactions now es)
@@ -118,8 +119,8 @@ tests_Hledger_Data_TimeLog = TestList [
      assertEntriesGiveStrings "split multi-day sessions at each midnight"
                                   [clockin (mktime (addDays (-2) today) "23:00:00") ""]
                                   ["23:00-23:59","00:00-23:59","00:00-"++nowstr]
-     assertEntriesGiveStrings "auto-clock-out if needed" 
-                                  [clockin (mktime today "00:00:00") ""] 
+     assertEntriesGiveStrings "auto-clock-out if needed"
+                                  [clockin (mktime today "00:00:00") ""]
                                   ["00:00-"++nowstr]
      let future = utcToLocalTime tz $ addUTCTime 100 now'
          futurestr = showtime future
@@ -127,4 +128,4 @@ tests_Hledger_Data_TimeLog = TestList [
                                   [clockin future ""]
                                   [printf "%s-%s" futurestr futurestr]
 
- ]
\ No newline at end of file
+ ]
