@@ -1,4 +1,4 @@
-{-# LANGUAGE DeriveDataTypeable, StandaloneDeriving #-}
+{-# LANGUAGE CPP, DeriveDataTypeable, StandaloneDeriving, TypeSynonymInstances, FlexibleInstances #-}
 {-|
 
 Most data types are defined here to avoid import cycles.
@@ -21,10 +21,17 @@ module Hledger.Data.Types
 where
 import Control.Monad.Error (ErrorT)
 import Data.Data
+#ifndef DOUBLE
+import Data.Decimal
+import Text.Blaze (ToMarkup(..))
+#endif
 import qualified Data.Map as M
 import Data.Time.Calendar
 import Data.Time.LocalTime
 import System.Time (ClockTime(..))
+import Text.Parsec.Pos
+
+import Hledger.Utils.Regex
 
 
 type SmartDate = (String,String,String)
@@ -41,11 +48,28 @@ data Interval = NoInterval
 
 type AccountName = String
 
+type AccountAlias = (Regexp,Replacement)
+
 data Side = L | R deriving (Eq,Show,Read,Ord,Typeable,Data)
 
 type Commodity = String
-      
+
+-- | The basic numeric type used in amounts. Different implementations
+-- can be selected via cabal flag for testing and benchmarking purposes.
+numberRepresentation :: String
+#ifdef DOUBLE
 type Quantity = Double
+numberRepresentation = "Double"
+#else
+type Quantity = Decimal
+deriving instance Data (Quantity)
+-- The following is for hledger-web, and requires blaze-markup.
+-- Doing it here avoids needing a matching flag on the hledger-web package.
+instance ToMarkup (Quantity) 
+ where
+   toMarkup = toMarkup . show
+numberRepresentation = "Decimal"
+#endif
 
 -- | An amount's price (none, per unit, or total) in another commodity.
 -- Note the price should be a positive number, although this is not enforced.
@@ -56,11 +80,19 @@ data AmountStyle = AmountStyle {
       ascommodityside :: Side,       -- ^ does the symbol appear on the left or the right ?
       ascommodityspaced :: Bool,     -- ^ space between symbol and quantity ?
       asprecision :: Int,            -- ^ number of digits displayed after the decimal point
-      asdecimalpoint :: Char,        -- ^ character used as decimal point
-      asseparator :: Char,           -- ^ character used for separating digit groups (eg thousands)
-      asseparatorpositions :: [Int]  -- ^ positions of digit group separators, counting leftward from decimal point
+      asdecimalpoint :: Maybe Char,  -- ^ character used as decimal point: period or comma. Nothing means "unspecified, use default"
+      asdigitgroups :: Maybe DigitGroupStyle -- ^ style for displaying digit groups, if any
 } deriving (Eq,Ord,Read,Show,Typeable,Data)
 
+-- | A style for displaying digit groups in the integer part of a
+-- floating point number. It consists of the character used to
+-- separate groups (comma or period, whichever is not used as decimal
+-- point), and the size of each group, starting with the one nearest
+-- the decimal point. The last group size is assumed to repeat. Eg,
+-- comma between thousands is DigitGroups ',' [3].
+data DigitGroupStyle = DigitGroups Char [Int]
+  deriving (Eq,Ord,Read,Show,Typeable,Data)
+
 data Amount = Amount {
       acommodity :: Commodity,
       aquantity :: Quantity,
@@ -95,6 +127,7 @@ instance Eq Posting where
     (==) (Posting a1 b1 c1 d1 e1 f1 g1 h1 i1 _) (Posting a2 b2 c2 d2 e2 f2 g2 h2 i2 _) =  a1==a2 && b1==b2 && c1==c2 && d1==d2 && e1==e2 && f1==f2 && g1==g2 && h1==h2 && i1==i2
 
 data Transaction = Transaction {
+      tsourcepos :: SourcePos,
       tdate :: Day,
       tdate2 :: Maybe Day,
       tstatus :: Bool,  -- XXX tcleared ?
@@ -119,6 +152,7 @@ data PeriodicTransaction = PeriodicTransaction {
 data TimeLogCode = SetBalance | SetRequiredHours | In | Out | FinalOut deriving (Eq,Ord,Typeable,Data)
 
 data TimeLogEntry = TimeLogEntry {
+      tlsourcepos :: SourcePos,
       tlcode :: TimeLogCode,
       tldatetime :: LocalTime,
       tlcomment :: String
@@ -138,11 +172,11 @@ type Year = Integer
 -- is saved for later use by eg the add command.
 data JournalContext = Ctx {
       ctxYear      :: !(Maybe Year)      -- ^ the default year most recently specified with Y
-    , ctxCommodityAndStyle :: !(Maybe (Commodity,AmountStyle)) -- ^ the default commodity and amount style most recently specified with D
+    , ctxDefaultCommodityAndStyle :: !(Maybe (Commodity,AmountStyle)) -- ^ the default commodity and amount style most recently specified with D
     , ctxAccount   :: ![AccountName]     -- ^ the current stack of parent accounts/account name components
                                         --   specified with "account" directive(s). Concatenated, these
                                         --   are the account prefix prepended to parsed account names.
-    , ctxAliases   :: ![(AccountName,AccountName)] -- ^ the current list of account name aliases in effect
+    , ctxAliases   :: ![AccountAlias]   -- ^ the current list of account name aliases in effect
     } deriving (Read, Show, Eq, Data, Typeable)
 
 deriving instance Data (ClockTime)
@@ -159,7 +193,7 @@ data Journal = Journal {
       files :: [(FilePath, String)],        -- ^ the file path and raw text of the main and
                                             -- any included journal files. The main file is
                                             -- first followed by any included files in the
-                                            -- order encountered (XXX reversed, cf journalAddFile).
+                                            -- order encountered.
       filereadtime :: ClockTime,            -- ^ when this journal was last read from its file(s)
       jcommoditystyles :: M.Map Commodity AmountStyle  -- ^ how to display amounts in each commodity
     } deriving (Eq, Typeable, Data)
@@ -179,10 +213,10 @@ data Reader = Reader {
      -- quickly check if this reader can probably handle the given file path and file content
     ,rDetector :: FilePath -> String -> Bool
      -- parse the given string, using the given parse rules file if any, returning a journal or error aware of the given file path
-    ,rParser   :: Maybe FilePath -> FilePath -> String -> ErrorT String IO Journal
+    ,rParser   :: Maybe FilePath -> Bool -> FilePath -> String -> ErrorT String IO Journal
     }
 
-instance Show Reader where show r = "Reader for "++rFormat r
+instance Show Reader where show r = rFormat r ++ " reader"
 
 -- format strings
 
