@@ -19,10 +19,14 @@ module Hledger.Query (
   queryIsAcct,
   queryIsDepth,
   queryIsDate,
+  queryIsDate2,
+  queryIsDateOrDate2,
   queryIsStartDateOnly,
   queryIsSym,
   queryStartDate,
+  queryEndDate,
   queryDateSpan,
+  queryDateSpan',
   queryDepth,
   queryEmpty,
   inAccount,
@@ -45,7 +49,8 @@ import Data.Maybe
 import Data.Time.Calendar
 import Safe (readDef, headDef, headMay)
 import Test.HUnit
-import Text.ParserCombinators.Parsec
+-- import Text.ParserCombinators.Parsec
+import Text.Parsec hiding (Empty)
 
 import Hledger.Utils
 import Hledger.Data.Types
@@ -206,7 +211,7 @@ prefixes = map (++":") [
     ,"desc"
     ,"acct"
     ,"date"
-    ,"edate"
+    ,"date2"
     ,"status"
     ,"cur"
     ,"real"
@@ -227,7 +232,8 @@ defaultprefix = "acct"
 -- query :: GenParser String () Query
 -- query = undefined
 
--- | Parse a single query term as either a query or a query option.
+-- | Parse a single query term as either a query or a query option,
+-- or raise an error if it has invalid syntax.
 parseQueryTerm :: Day -> String -> Either Query QueryOpt
 parseQueryTerm _ ('i':'n':'a':'c':'c':'t':'o':'n':'l':'y':':':s) = Right $ QueryOptInAcctOnly s
 parseQueryTerm _ ('i':'n':'a':'c':'c':'t':':':s) = Right $ QueryOptInAcct s
@@ -237,12 +243,12 @@ parseQueryTerm d ('n':'o':'t':':':s) = case parseQueryTerm d s of
 parseQueryTerm _ ('c':'o':'d':'e':':':s) = Left $ Code s
 parseQueryTerm _ ('d':'e':'s':'c':':':s) = Left $ Desc s
 parseQueryTerm _ ('a':'c':'c':'t':':':s) = Left $ Acct s
-parseQueryTerm d ('d':'a':'t':'e':':':s) =
-        case parsePeriodExpr d s of Left _ -> Left None -- XXX should warn
-                                    Right (_,span) -> Left $ Date span
-parseQueryTerm d ('e':'d':'a':'t':'e':':':s) =
-        case parsePeriodExpr d s of Left _ -> Left None -- XXX should warn
+parseQueryTerm d ('d':'a':'t':'e':'2':':':s) =
+        case parsePeriodExpr d s of Left e         -> error' $ "\"date2:"++s++"\" gave a "++showDateParseError e
                                     Right (_,span) -> Left $ Date2 span
+parseQueryTerm d ('d':'a':'t':'e':':':s) =
+        case parsePeriodExpr d s of Left e         -> error' $ "\"date:"++s++"\" gave a "++showDateParseError e
+                                    Right (_,span) -> Left $ Date span
 parseQueryTerm _ ('s':'t':'a':'t':'u':'s':':':s) = Left $ Status $ parseStatus s
 parseQueryTerm _ ('r':'e':'a':'l':':':s) = Left $ Real $ parseBool s
 parseQueryTerm _ ('a':'m':'t':':':s) = Left $ Amt ord q where (ord, q) = parseAmountQueryTerm s
@@ -274,7 +280,7 @@ tests_parseQueryTerm = [
  ]
 
 
-data OrdPlus = Lt | Gt | Eq | AbsLt | AbsGt | AbsEq
+data OrdPlus = Lt | LtEq | Gt | GtEq | Eq | AbsLt | AbsLtEq | AbsGt | AbsGtEq | AbsEq
  deriving (Show,Eq,Data,Typeable)
 
 -- can fail
@@ -282,21 +288,29 @@ parseAmountQueryTerm :: String -> (OrdPlus, Quantity)
 parseAmountQueryTerm s' =
   case s' of
     -- feel free to do this a smarter way
-    ""        -> err
-    '<':'+':s -> (Lt, readDef err s)
-    '>':'+':s -> (Gt, readDef err s)
-    '=':'+':s -> (Eq, readDef err s)
-    '+':s     -> (Eq, readDef err s)
-    '<':'-':s -> (Lt, negate $ readDef err s)
-    '>':'-':s -> (Gt, negate $ readDef err s)
-    '=':'-':s -> (Eq, negate $ readDef err s)
-    '-':s     -> (Eq, negate $ readDef err s)
-    '<':s     -> let n = readDef err s in case n of 0 -> (Lt, 0)
-                                                    _ -> (AbsLt, n)
-    '>':s     -> let n = readDef err s in case n of 0 -> (Gt, 0)
-                                                    _ -> (AbsGt, n)
-    '=':s     -> (AbsEq, readDef err s)
-    s         -> (AbsEq, readDef err s)
+    ""              -> err
+    '<':'+':s       -> (Lt, readDef err s)
+    '<':'=':'+':s   -> (LtEq, readDef err s)
+    '>':'+':s       -> (Gt, readDef err s)
+    '>':'=':'+':s   -> (GtEq, readDef err s)
+    '=':'+':s       -> (Eq, readDef err s)
+    '+':s           -> (Eq, readDef err s)
+    '<':'-':s       -> (Lt, negate $ readDef err s)
+    '<':'=':'-':s   -> (LtEq, negate $ readDef err s)
+    '>':'-':s       -> (Gt, negate $ readDef err s)
+    '>':'=':'-':s   -> (GtEq, negate $ readDef err s)
+    '=':'-':s       -> (Eq, negate $ readDef err s)
+    '-':s           -> (Eq, negate $ readDef err s)
+    '<':'=':s       -> let n = readDef err s in case n of 0 -> (LtEq, 0)
+                                                          _ -> (AbsLtEq, n)
+    '<':s           -> let n = readDef err s in case n of 0 -> (Lt, 0)
+                                                          _ -> (AbsLt, n)
+    '>':'=':s       -> let n = readDef err s in case n of 0 -> (GtEq, 0)
+                                                          _ -> (AbsGtEq, n)
+    '>':s           -> let n = readDef err s in case n of 0 -> (Gt, 0)
+                                                          _ -> (AbsGt, n)
+    '=':s           -> (AbsEq, readDef err s)
+    s               -> (AbsEq, readDef err s)
   where
     err = error' $ "could not parse as '=', '<', or '>' (optional) followed by a (optionally signed) numeric quantity: " ++ s'
 
@@ -308,7 +322,7 @@ tests_parseAmountQueryTerm = [
     ">10000.10" `gives` (AbsGt,10000.1)
     "=0.23" `gives` (AbsEq,0.23)
     "0.23" `gives` (AbsEq,0.23)
-    "=+0.23" `gives` (Eq,0.23)
+    "<=+0.23" `gives` (LtEq,0.23)
     "-0.23" `gives` (Eq,(-0.23))
   ]
 
@@ -317,18 +331,18 @@ parseTag s | '=' `elem` s = (n, Just $ tail v)
            | otherwise    = (s, Nothing)
            where (n,v) = break (=='=') s
 
--- | Parse the boolean value part of a "status:" query, allowing "*" as
--- another way to spell True, similar to the journal file format.
+-- -- , treating "*" or "!" as synonyms for "1".
+-- | Parse the boolean value part of a "status:" query.
 parseStatus :: String -> Bool
-parseStatus s = s `elem` (truestrings ++ ["*"])
+parseStatus s = s `elem` (truestrings) -- ++ ["*","!"])
 
--- | Parse the boolean value part of a "status:" query. A true value can
--- be spelled as "1", "t" or "true".
+-- | Parse the boolean value part of a "status:" query. "1" means true,
+-- anything else will be parsed as false without error.
 parseBool :: String -> Bool
 parseBool s = s `elem` truestrings
 
 truestrings :: [String]
-truestrings = ["1","t","true"]
+truestrings = ["1"]
 
 simplifyQuery :: Query -> Query
 simplifyQuery q =
@@ -349,6 +363,7 @@ simplifyQuery q =
                      -- all queryIsDate qs = Date $ spansUnion $ mapMaybe queryTermDateSpan qs  ?
                      | otherwise = Or $ map simplify $ filter (/=None) qs
     simplify (Date (DateSpan Nothing Nothing)) = Any
+    simplify (Date2 (DateSpan Nothing Nothing)) = Any
     simplify q = q
 
 tests_simplifyQuery = [
@@ -403,9 +418,17 @@ queryIsDepth _ = False
 
 queryIsDate :: Query -> Bool
 queryIsDate (Date _) = True
-queryIsDate (Date2 _) = True
 queryIsDate _ = False
 
+queryIsDate2 :: Query -> Bool
+queryIsDate2 (Date2 _) = True
+queryIsDate2 _ = False
+
+queryIsDateOrDate2 :: Query -> Bool
+queryIsDateOrDate2 (Date _) = True
+queryIsDateOrDate2 (Date2 _) = True
+queryIsDateOrDate2 _ = False
+
 queryIsDesc :: Query -> Bool
 queryIsDesc (Desc _) = True
 queryIsDesc _ = False
@@ -439,6 +462,15 @@ queryStartDate False (Date (DateSpan (Just d) _)) = Just d
 queryStartDate True (Date2 (DateSpan (Just d) _)) = Just d
 queryStartDate _ _ = Nothing
 
+-- | What end date (or secondary date) does this query specify, if any ?
+-- For OR expressions, use the latest of the dates. NOT is ignored.
+queryEndDate :: Bool -> Query -> Maybe Day
+queryEndDate secondary (Or ms) = latestMaybeDate' $ map (queryEndDate secondary) ms
+queryEndDate secondary (And ms) = earliestMaybeDate' $ map (queryEndDate secondary) ms
+queryEndDate False (Date (DateSpan _ (Just d))) = Just d
+queryEndDate True (Date2 (DateSpan _ (Just d))) = Just d
+queryEndDate _ _ = Nothing
+
 queryTermDateSpan (Date span) = Just span
 queryTermDateSpan _ = Nothing
 
@@ -456,14 +488,36 @@ queryDateSpans False (Date span) = [span]
 queryDateSpans True (Date2 span) = [span]
 queryDateSpans _ _ = []
 
--- | What is the earliest of these dates, where Nothing is earliest ?
+-- | What date span (or secondary date span) does this query specify ?
+-- For OR expressions, use the widest possible span. NOT is ignored.
+queryDateSpan' :: Query -> DateSpan
+queryDateSpan' q = spansUnion $ queryDateSpans' q
+
+-- | Extract all date (or secondary date) spans specified in this query.
+-- NOT is ignored.
+queryDateSpans' :: Query -> [DateSpan]
+queryDateSpans' (Or qs) = concatMap queryDateSpans' qs
+queryDateSpans' (And qs) = concatMap queryDateSpans' qs
+queryDateSpans' (Date span) = [span]
+queryDateSpans' (Date2 span) = [span]
+queryDateSpans' _ = []
+
+-- | What is the earliest of these dates, where Nothing is latest ?
 earliestMaybeDate :: [Maybe Day] -> Maybe Day
-earliestMaybeDate = headDef Nothing . sortBy compareMaybeDates
+earliestMaybeDate mds = head $ sortBy compareMaybeDates mds ++ [Nothing]
 
 -- | What is the latest of these dates, where Nothing is earliest ?
 latestMaybeDate :: [Maybe Day] -> Maybe Day
 latestMaybeDate = headDef Nothing . sortBy (flip compareMaybeDates)
 
+-- | What is the earliest of these dates, ignoring Nothings ?
+earliestMaybeDate' :: [Maybe Day] -> Maybe Day
+earliestMaybeDate' = headDef Nothing . sortBy compareMaybeDates . filter isJust
+
+-- | What is the latest of these dates, ignoring Nothings ?
+latestMaybeDate' :: [Maybe Day] -> Maybe Day
+latestMaybeDate' = headDef Nothing . sortBy (flip compareMaybeDates) . filter isJust
+
 -- | Compare two maybe dates, Nothing is earliest.
 compareMaybeDates :: Maybe Day -> Maybe Day -> Ordering
 compareMaybeDates Nothing Nothing = EQ
@@ -566,12 +620,16 @@ matchesAmount _ _ = True
 
 -- | Is this amount's quantity less than, greater than, equal to, or unsignedly equal to this number ?
 compareAmount :: OrdPlus -> Quantity -> Amount -> Bool
-compareAmount ord q Amount{aquantity=aq} = case ord of Lt    -> aq <  q
-                                                       Gt    -> aq >  q
-                                                       Eq    -> aq == q
-                                                       AbsLt -> abs aq <  abs q
-                                                       AbsGt -> abs aq >  abs q
-                                                       AbsEq -> abs aq == abs q
+compareAmount ord q Amount{aquantity=aq} = case ord of Lt      -> aq <  q
+                                                       LtEq    -> aq <= q
+                                                       Gt      -> aq >  q
+                                                       GtEq    -> aq >= q
+                                                       Eq      -> aq == q
+                                                       AbsLt   -> abs aq <  abs q
+                                                       AbsLtEq -> abs aq <= abs q
+                                                       AbsGt   -> abs aq >  abs q
+                                                       AbsGtEq -> abs aq >= abs q
+                                                       AbsEq   -> abs aq == abs q
 
 -- | Does the match expression match this posting ?
 matchesPosting :: Query -> Posting -> Bool
