@@ -1,5 +1,5 @@
 {-# LANGUAGE ScopedTypeVariables #-}
-{-| 
+{-|
 
 This is the entry point to hledger's reading system, which can read
 Journals from various data formats. Use this module if you want to parse
@@ -91,11 +91,11 @@ defaultJournalPath = do
 
 -- | Read the default journal file specified by the environment, or raise an error.
 defaultJournal :: IO Journal
-defaultJournal = defaultJournalPath >>= readJournalFile Nothing Nothing >>= either error' return
+defaultJournal = defaultJournalPath >>= readJournalFile Nothing Nothing True >>= either error' return
 
 -- | Read a journal from the given string, trying all known formats, or simply throw an error.
 readJournal' :: String -> IO Journal
-readJournal' s = readJournal Nothing Nothing Nothing s >>= either error' return
+readJournal' s = readJournal Nothing Nothing True Nothing s >>= either error' return
 
 tests_readJournal' = [
   "readJournal' parses sample journal" ~: do
@@ -114,9 +114,9 @@ tests_readJournal' = [
 -- - otherwise, try them all.
 --
 -- A CSV conversion rules file may also be specified for use by the CSV reader.
-readJournal :: Maybe StorageFormat -> Maybe FilePath -> Maybe FilePath -> String -> IO (Either String Journal)
-readJournal format rulesfile path s =
-  -- trace (show (format, rulesfile, path)) $
+-- Also there is a flag specifying whether to check or ignore balance assertions in the journal.
+readJournal :: Maybe StorageFormat -> Maybe FilePath -> Bool -> Maybe FilePath -> String -> IO (Either String Journal)
+readJournal format rulesfile assrt path s =
   tryReaders $ readersFor (format, path, s)
   where
     -- try each reader in turn, returning the error of the first if all fail
@@ -126,8 +126,9 @@ readJournal format rulesfile path s =
         firstSuccessOrBestError :: [String] -> [Reader] -> IO (Either String Journal)
         firstSuccessOrBestError [] []        = return $ Left "no readers found"
         firstSuccessOrBestError errs (r:rs) = do
-          -- printf "trying %s reader\n" (rFormat r)
-          result <- (runErrorT . (rParser r) rulesfile path') s
+          dbgAtM 1 "trying reader" (rFormat r)
+          result <- (runErrorT . (rParser r) rulesfile assrt path') s
+          dbgAtM 1 "reader result" $ either id show result
           case result of Right j -> return $ Right j                       -- success!
                          Left e  -> firstSuccessOrBestError (errs++[e]) rs -- keep trying
         firstSuccessOrBestError (e:_) []    = return $ Left e              -- none left, return first error
@@ -136,11 +137,11 @@ readJournal format rulesfile path s =
 -- | Which readers are worth trying for this (possibly unspecified) format, filepath, and data ?
 readersFor :: (Maybe StorageFormat, Maybe FilePath, String) -> [Reader]
 readersFor (format,path,s) =
-    case format of 
+    dbg ("possible readers for "++show (format,path,elideRight 30 s)) $
+    case format of
      Just f  -> case readerForStorageFormat f of Just r  -> [r]
                                                  Nothing -> []
      Nothing -> case path of Nothing  -> readers
-                             Just "-" -> readers
                              Just p   -> case readersForPathAndData (p,s) of [] -> readers
                                                                              rs -> rs
 
@@ -148,7 +149,7 @@ readersFor (format,path,s) =
 readerForStorageFormat :: StorageFormat -> Maybe Reader
 readerForStorageFormat s | null rs = Nothing
                   | otherwise = Just $ head rs
-    where 
+    where
       rs = filter ((s==).rFormat) readers :: [Reader]
 
 -- | Find the readers which think they can handle the given file path and data, if any.
@@ -158,16 +159,17 @@ readersForPathAndData (f,s) = filter (\r -> (rDetector r) f s) readers
 -- | Read a Journal from this file (or stdin if the filename is -) or give
 -- an error message, using the specified data format or trying all known
 -- formats. A CSV conversion rules file may be specified for better
--- conversion of that format.
-readJournalFile :: Maybe StorageFormat -> Maybe FilePath -> FilePath -> IO (Either String Journal)
-readJournalFile format rulesfile "-" = do
+-- conversion of that format. Also there is a flag specifying whether
+-- to check or ignore balance assertions in the journal.
+readJournalFile :: Maybe StorageFormat -> Maybe FilePath -> Bool -> FilePath -> IO (Either String Journal)
+readJournalFile format rulesfile assrt "-" = do
   hSetNewlineMode stdin universalNewlineMode
-  getContents >>= readJournal format rulesfile (Just "(stdin)")
-readJournalFile format rulesfile f = do
+  getContents >>= readJournal format rulesfile assrt (Just "-")
+readJournalFile format rulesfile assrt f = do
   requireJournalFileExists f
   withFile f ReadMode $ \h -> do
     hSetNewlineMode h universalNewlineMode
-    hGetContents h >>= readJournal format rulesfile (Just f)
+    hGetContents h >>= readJournal format rulesfile assrt (Just f)
 
 -- | If the specified journal file does not exist, give a helpful error and quit.
 requireJournalFileExists :: FilePath -> IO ()
@@ -202,6 +204,11 @@ samplejournal = readJournal' $ unlines
  ,"    assets:bank:checking  $1"
  ,"    income:salary"
  ,""
+ ,"comment"
+ ,"multi line comment here"
+ ,"for testing purposes"
+ ,"end comment"
+ ,""
  ,"2008/06/01 gift"
  ,"    assets:bank:checking  $1"
  ,"    income:gifts"
@@ -228,8 +235,9 @@ tests_Hledger_Read = TestList $
    tests_Hledger_Read_CsvReader,
 
    "journal" ~: do
-    assertBool "journal should parse an empty file" (isRight $ parseWithCtx nullctx JournalReader.journal "")
-    jE <- readJournal Nothing Nothing Nothing "" -- don't know how to get it from journal
+    r <- runErrorT $ parseWithCtx nullctx JournalReader.journal ""
+    assertBool "journal should parse an empty file" (isRight $ r)
+    jE <- readJournal Nothing Nothing True Nothing "" -- don't know how to get it from journal
     either error' (assertBool "journal parsing an empty file should give an empty journal" . null . jtxns) jE
 
   ]
