@@ -3,6 +3,8 @@
 A reader for CSV data, using an extra rules file to help interpret the data.
 
 -}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE ScopedTypeVariables #-}
 
 module Hledger.Read.CsvReader (
   -- * Reader
@@ -17,7 +19,7 @@ module Hledger.Read.CsvReader (
   tests_Hledger_Read_CsvReader
 )
 where
-import Control.Applicative ((<$>))
+import Control.Applicative ((<$>), (<*))
 import Control.Exception hiding (try)
 import Control.Monad
 import Control.Monad.Error
@@ -35,9 +37,9 @@ import System.IO (stderr)
 import System.Locale (defaultTimeLocale)
 import Test.HUnit
 import Text.CSV (parseCSV, CSV)
-import Text.ParserCombinators.Parsec  hiding (parse)
-import Text.ParserCombinators.Parsec.Error
-import Text.ParserCombinators.Parsec.Pos
+import Text.Parsec hiding (parse)
+import Text.Parsec.Pos
+import Text.Parsec.Error
 import Text.Printf (hPrintf,printf)
 
 import Hledger.Data
@@ -53,15 +55,16 @@ reader = Reader format detect parse
 format :: String
 format = "csv"
 
--- | Does the given file path and data look like CSV ?
+-- | Does the given file path and data look like it might be CSV ?
 detect :: FilePath -> String -> Bool
-detect f _ = takeExtension f == '.':format
+detect f s
+  | f /= "-"  = takeExtension f == '.':format  -- from a file: yes if the extension is .csv
+  | otherwise = length (filter (==',') s) >= 2 -- from stdin: yes if there are two or more commas
 
 -- | Parse and post-process a "Journal" from CSV data, or give an error.
 -- XXX currently ignores the string and reads from the file path
-parse :: Maybe FilePath -> FilePath -> String -> ErrorT String IO Journal
-parse rulesfile f s = -- trace ("running "++format++" reader") $
- do
+parse :: Maybe FilePath -> Bool -> FilePath -> String -> ErrorT String IO Journal
+parse rulesfile _ f s = do
   r <- liftIO $ readJournalFromCsv rulesfile f s
   case r of Left e -> throwError e
             Right j -> return j
@@ -78,7 +81,7 @@ parse rulesfile f s = -- trace ("running "++format++" reader") $
 -- 5. convert the CSV records to a journal using the rules
 -- @
 readJournalFromCsv :: Maybe FilePath -> FilePath -> String -> IO (Either String Journal)
-readJournalFromCsv Nothing "-" _ = return $ Left "please use --rules-file when converting stdin"
+readJournalFromCsv Nothing "-" _ = return $ Left "please use --rules-file when reading CSV from stdin"
 readJournalFromCsv mrulesfile csvfile csvdata =
  handle (\e -> return $ Left $ show (e :: IOException)) $ do
   let throwerr = throw.userError
@@ -89,8 +92,11 @@ readJournalFromCsv mrulesfile csvfile csvdata =
   if created
    then hPrintf stderr "creating default conversion rules file %s, edit this file for better results\n" rulesfile
    else hPrintf stderr "using conversion rules file %s\n" rulesfile
-  rules <- either (throwerr.show) id `fmap` parseRulesFile rulesfile
-  return $ dbg "" rules
+  rules_ <- liftIO $ runErrorT $ parseRulesFile rulesfile
+  let rules = case rules_ of
+              Right (t::CsvRules) -> t
+              Left err -> throwerr $ show err
+  dbgAtM 2 "rules" rules
 
   -- apply skip directive
   let skip = maybe 0 oneorerror $ getDirective "skip" rules
@@ -99,16 +105,29 @@ readJournalFromCsv mrulesfile csvfile csvdata =
           oneorerror s  = readDef (throwerr $ "could not parse skip value: " ++ show s) s
 
   -- parse csv
-  records <- (either throwerr id . validateCsv skip) `fmap` parseCsv csvfile csvdata
-  dbgAtM 1 "" $ take 3 records
+  -- parsec seems to fail if you pass it "-" here
+  let parsecfilename = if csvfile == "-" then "(stdin)" else csvfile
+  records <- (either throwerr id .
+              dbgAt 2 "validateCsv" . validateCsv skip .
+              dbgAt 2 "parseCsv")
+             `fmap` parseCsv parsecfilename csvdata
+  dbgAtM 1 "first 3 csv records" $ take 3 records
 
   -- identify header lines
   -- let (headerlines, datalines) = identifyHeaderLines records
   --     mfieldnames = lastMay headerlines
 
   -- convert to transactions and return as a journal
-  let txns = map (transactionFromCsvRecord rules) records
-  return $ Right nulljournal{jtxns=sortBy (comparing tdate) txns}
+  let txns = snd $ mapAccumL
+                     (\pos r -> (pos, transactionFromCsvRecord (incSourceLine pos 1) rules r))
+                     (initialPos parsecfilename) records
+
+  -- heuristic: if the records appear to have been in reverse date order,
+  -- reverse them all as well as doing a txn date sort,
+  -- so that same-day txns' original order is preserved
+      txns' | length txns > 1 && tdate (head txns) > tdate (last txns) = reverse txns
+            | otherwise = txns
+  return $ Right nulljournal{jtxns=sortBy (comparing tdate) txns'}
 
 parseCsv :: FilePath -> String -> IO (Either ParseError CSV)
 parseCsv path csvdata =
@@ -173,7 +192,7 @@ ensureRulesFileExists f = do
 newRulesFileContent :: FilePath -> String
 newRulesFileContent f = unlines
   ["# hledger csv conversion rules for " ++ csvFileFor (takeFileName f)
-  ,"# cf http://hledger.org/MANUAL.html"
+  ,"# cf http://hledger.org/manual#csv-files"
   ,""
   ,"account1 assets:bank:checking"
   ,""
@@ -278,10 +297,10 @@ type CsvFieldIndex    = Int
 type JournalFieldName = String
 type FieldTemplate    = String
 type ConditionalBlock = ([RecordMatcher], [(JournalFieldName, FieldTemplate)]) -- block matches if all RecordMatchers match
-type RecordMatcher    = [Regexp] -- match if any regexps match any of the csv fields
--- type FieldMatcher     = (CsvFieldName, [Regexp]) -- match if any regexps match this csv field
+type RecordMatcher    = [RegexpPattern] -- match if any regexps match any of the csv fields
+-- type FieldMatcher     = (CsvFieldName, [RegexpPattern]) -- match if any regexps match this csv field
 type DateFormat       = String
-type Regexp           = String
+type RegexpPattern           = String
 
 rules = CsvRules {
   rdirectives=[],
@@ -316,28 +335,31 @@ getDirective :: DirectiveName -> CsvRules -> Maybe FieldTemplate
 getDirective directivename = lookup directivename . rdirectives
 
 
-parseRulesFile :: FilePath -> IO (Either ParseError CsvRules)
+parseRulesFile :: FilePath -> ErrorT String IO CsvRules
 parseRulesFile f = do
-  s <- readFile' f >>= expandIncludes
+  s <- liftIO $ (readFile' f >>= expandIncludes (takeDirectory f))
   let rules = parseCsvRules f s
-  return $ case rules of
-             Left e -> Left e
-             Right r -> case validateRules r of
-                          Left e -> Left $ toParseError e
-                          Right r -> Right r
+  case rules of
+    Left e -> ErrorT $ return $ Left $ show e
+    Right r -> do
+               r_ <- liftIO $ runErrorT $ validateRules r
+               ErrorT $ case r_ of
+                 Left e -> return $ Left $ show $ toParseError e
+                 Right r -> return $ Right r
   where
     toParseError s = newErrorMessage (Message s) (initialPos "")
 
 -- | Pre-parse csv rules to interpolate included files, recursively.
 -- This is a cheap hack to avoid rewriting the existing parser.
-expandIncludes :: String -> IO String
-expandIncludes s = do
-  let (ls,rest) = break (isPrefixOf "include") $ lines s
+expandIncludes :: FilePath -> String -> IO String
+expandIncludes basedir content = do
+  let (ls,rest) = break (isPrefixOf "include") $ lines content
   case rest of
     [] -> return $ unlines ls
     (('i':'n':'c':'l':'u':'d':'e':f):ls') -> do
-      let f' = dropWhile isSpace f
-      included <- readFile f' >>= expandIncludes
+      let f'       = basedir </> dropWhile isSpace f
+          basedir' = takeDirectory f'
+      included <- readFile f' >>= expandIncludes basedir'
       return $ unlines [unlines ls, included, unlines ls']
     ls' -> return $ unlines $ ls ++ ls'   -- should never get here
 
@@ -347,13 +369,13 @@ parseCsvRules rulesfile s =
   runParser rulesp rules rulesfile s
 
 -- | Return the validated rules, or an error.
-validateRules :: CsvRules -> Either String CsvRules
+validateRules :: CsvRules -> ErrorT String IO CsvRules
 validateRules rules = do
-  unless (isAssigned "date")   $ Left "Please specify (at top level) the date field. Eg: date %1\n"
+  unless (isAssigned "date")   $ ErrorT $ return $ Left "Please specify (at top level) the date field. Eg: date %1\n"
   unless ((amount && not (amountin || amountout)) ||
           (not amount && (amountin && amountout)))
-    $ Left "Please specify (at top level) either the amount field, or both the amount-in and amount-out fields. Eg: amount %2\n"
-  Right rules
+    $ ErrorT $ return $ Left "Please specify (at top level) either the amount field, or both the amount-in and amount-out fields. Eg: amount %2\n"
+  ErrorT $ return $ Right rules
   where
     amount = isAssigned "amount"
     amountin = isAssigned "amount-in"
@@ -362,14 +384,14 @@ validateRules rules = do
 
 -- parsers
 
-rulesp :: GenParser Char CsvRules CsvRules
+rulesp :: Stream [Char] m t => ParsecT [Char] CsvRules m CsvRules
 rulesp = do
   many $ choice'
     [blankorcommentline                                                    <?> "blank or comment line"
-    ,(directive        >>= updateState . addDirective)                     <?> "directive"
-    ,(fieldnamelist    >>= updateState . setIndexesAndAssignmentsFromList) <?> "field name list"
-    ,(fieldassignment  >>= updateState . addAssignment)                    <?> "field assignment"
-    ,(conditionalblock >>= updateState . addConditionalBlock)              <?> "conditional block"
+    ,(directive        >>= modifyState . addDirective)                     <?> "directive"
+    ,(fieldnamelist    >>= modifyState . setIndexesAndAssignmentsFromList) <?> "field name list"
+    ,(fieldassignment  >>= modifyState . addAssignment)                    <?> "field assignment"
+    ,(conditionalblock >>= modifyState . addConditionalBlock)              <?> "conditional block"
     ]
   eof
   r <- getState
@@ -378,13 +400,21 @@ rulesp = do
           ,rconditionalblocks=reverse $ rconditionalblocks r
           }
 
-blankorcommentline = pdbg 1 "trying blankorcommentline" >> choice' [blankline, commentline]
+blankorcommentline :: Stream [Char] m t => ParsecT [Char] CsvRules m ()
+blankorcommentline = pdbg 3 "trying blankorcommentline" >> choice' [blankline, commentline]
+
+blankline :: Stream [Char] m t => ParsecT [Char] CsvRules m ()
 blankline = many spacenonewline >> newline >> return () <?> "blank line"
+
+commentline :: Stream [Char] m t => ParsecT [Char] CsvRules m ()
 commentline = many spacenonewline >> commentchar >> restofline >> return () <?> "comment line"
+
+commentchar :: Stream [Char] m t => ParsecT [Char] CsvRules m Char
 commentchar = oneOf ";#"
 
+directive :: Stream [Char] m t => ParsecT [Char] CsvRules m (DirectiveName, String)
 directive = do
-  pdbg 1 "trying directive"
+  pdbg 3 "trying directive"
   d <- choice' $ map string directives
   v <- (((char ':' >> many spacenonewline) <|> many1 spacenonewline) >> directiveval)
        <|> (optional (char ':') >> many spacenonewline >> eolof >> return "")
@@ -401,10 +431,12 @@ directives =
    -- ,"base-currency"
   ]
 
+directiveval :: Stream [Char] m t => ParsecT [Char] CsvRules m [Char]
 directiveval = anyChar `manyTill` eolof
 
+fieldnamelist :: Stream [Char] m t => ParsecT [Char] CsvRules m [CsvFieldName]
 fieldnamelist = (do
-  pdbg 1 "trying fieldnamelist"
+  pdbg 3 "trying fieldnamelist"
   string "fields"
   optional $ char ':'
   many1 spacenonewline
@@ -415,24 +447,29 @@ fieldnamelist = (do
   return $ map (map toLower) $ f:fs
   ) <?> "field name list"
 
+fieldname :: Stream [Char] m t => ParsecT [Char] CsvRules m [Char]
 fieldname = quotedfieldname <|> barefieldname
 
+quotedfieldname :: Stream [Char] m t => ParsecT [Char] CsvRules m [Char]
 quotedfieldname = do
   char '"'
   f <- many1 $ noneOf "\"\n:;#~"
   char '"'
   return f
 
+barefieldname :: Stream [Char] m t => ParsecT [Char] CsvRules m [Char]
 barefieldname = many1 $ noneOf " \t\n,;#~"
 
+fieldassignment :: Stream [Char] m t => ParsecT [Char] CsvRules m (JournalFieldName, FieldTemplate)
 fieldassignment = do
-  pdbg 1 "trying fieldassignment"
+  pdbg 3 "trying fieldassignment"
   f <- journalfieldname
   assignmentseparator
   v <- fieldval
   return (f,v)
   <?> "field assignment"
 
+journalfieldname :: Stream [Char] m t => ParsecT [Char] CsvRules m [Char]
 journalfieldname = pdbg 2 "trying journalfieldname" >> choice' (map string journalfieldnames)
 
 journalfieldnames =
@@ -452,6 +489,7 @@ journalfieldnames =
   ,"comment"
   ]
 
+assignmentseparator :: Stream [Char] m t => ParsecT [Char] CsvRules m ()
 assignmentseparator = do
   pdbg 3 "trying assignmentseparator"
   choice [
@@ -459,14 +497,17 @@ assignmentseparator = do
     try (many spacenonewline >> char ':'),
     space
     ]
-  many spacenonewline
+  _ <- many spacenonewline
+  return ()
 
+fieldval :: Stream [Char] m t => ParsecT [Char] CsvRules m [Char]
 fieldval = do
   pdbg 2 "trying fieldval"
   anyChar `manyTill` eolof
 
+conditionalblock :: Stream [Char] m t => ParsecT [Char] CsvRules m ConditionalBlock
 conditionalblock = do
-  pdbg 1 "trying conditionalblock"
+  pdbg 3 "trying conditionalblock"
   string "if" >> many spacenonewline >> optional newline
   ms <- many1 recordmatcher
   as <- many (many1 spacenonewline >> fieldassignment)
@@ -475,6 +516,7 @@ conditionalblock = do
   return (ms, as)
   <?> "conditional block"
 
+recordmatcher :: Stream [Char] m t => ParsecT [Char] CsvRules m [[Char]]
 recordmatcher = do
   pdbg 2 "trying recordmatcher"
   -- pos <- currentPos
@@ -485,6 +527,7 @@ recordmatcher = do
   return ps
   <?> "record matcher"
 
+matchoperator :: Stream [Char] m t => ParsecT [Char] CsvRules m [Char]
 matchoperator = choice' $ map string
   ["~"
   -- ,"!~"
@@ -492,11 +535,13 @@ matchoperator = choice' $ map string
   -- ,"!="
   ]
 
+patterns :: Stream [Char] m t => ParsecT [Char] CsvRules m [[Char]]
 patterns = do
   pdbg 3 "trying patterns"
   ps <- many regexp
   return ps
 
+regexp :: Stream [Char] m t => ParsecT [Char] CsvRules m [Char]
 regexp = do
   pdbg 3 "trying regexp"
   notFollowedBy matchoperator
@@ -524,8 +569,8 @@ type CsvRecord = [String]
 
 -- Convert a CSV record to a transaction using the rules, or raise an
 -- error if the data can not be parsed.
-transactionFromCsvRecord :: CsvRules -> CsvRecord -> Transaction
-transactionFromCsvRecord rules record = t
+transactionFromCsvRecord :: SourcePos -> CsvRules -> CsvRecord -> Transaction
+transactionFromCsvRecord sourcepos rules record = t
   where
     mdirective       = (`getDirective` rules)
     mfieldtemplate   = getEffectiveAssignment rules record
@@ -559,7 +604,7 @@ transactionFromCsvRecord rules record = t
     precomment  = maybe "" render $ mfieldtemplate "precomment"
     currency    = maybe (fromMaybe "" mdefaultcurrency) render $ mfieldtemplate "currency"
     amountstr   = (currency++) $ negateIfParenthesised $ getAmountStr rules record
-    amount      = either amounterror mixed $ runParser (do {a <- amountp; eof; return a}) nullctx "" amountstr
+    amount      = either amounterror (Mixed . (:[])) $ runParser (amountp <* eof) nullctx "" amountstr
     amounterror err = error' $ unlines
       ["error: could not parse \""++amountstr++"\" as an amount"
       ,showRecord record
@@ -585,6 +630,7 @@ transactionFromCsvRecord rules record = t
 
     -- build the transaction
     t = nulltransaction{
+      tsourcepos               = sourcepos,
       tdate                    = date',
       tdate2                   = mdate2',
       tstatus                  = status,
@@ -648,8 +694,8 @@ getEffectiveAssignment rules record f = lastMay $ assignmentsFor f
                 -- matcherMatches pats = any patternMatches pats
                 matcherMatches pats = patternMatches $  "(" ++ intercalate "|" pats ++ ")"
                   where
-                    patternMatches :: Regexp -> Bool
-                    patternMatches pat = regexMatchesCIRegexCompat pat csvline
+                    patternMatches :: RegexpPattern -> Bool
+                    patternMatches pat = regexMatchesCI pat csvline
                       where
                         csvline = intercalate "," record
 
