@@ -39,9 +39,9 @@ type BalanceReportItem = (RenderableAccountName, MixedAmount)
 -- It has:
 --
 -- * The full account name
--- 
+--
 -- * The ledger-style short elided account name (the leaf name, prefixed by any boring parents immediately above)
--- 
+--
 -- * The number of indentation steps to use when rendering a ledger-style account tree
 --   (normally the 0-based depth of this account excluding boring parents, or 0 with --flat).
 type RenderableAccountName = (AccountName, AccountName, Int)
@@ -67,15 +67,19 @@ balanceReport opts q j = (items, total)
 
       accts = ledgerRootAccount $ ledgerFromJournal q $ journalSelectingAmountFromOpts opts j
       accts' :: [Account]
-          | flat_ opts = dbg "accts" $ 
+          | queryDepth q == 0 =
+                         dbg "accts" $
+                         take 1 $ clipAccountsAndAggregate (queryDepth q) $ flattenAccounts accts
+          | flat_ opts = dbg "accts" $
                          filterzeros $
                          filterempty $
                          drop 1 $ clipAccountsAndAggregate (queryDepth q) $ flattenAccounts accts
-          | otherwise  = dbg "accts" $ 
+          | otherwise  = dbg "accts" $
                          filter (not.aboring) $
                          drop 1 $ flattenAccounts $
-                         markboring $ 
-                         prunezeros $ clipAccounts (queryDepth q) accts
+                         markboring $
+                         prunezeros $
+                         clipAccounts (queryDepth q) accts
           where
             balance     = if flat_ opts then aebalance else aibalance
             filterzeros = if empty_ opts then id else filter (not . isZeroMixedAmount . balance)
@@ -99,14 +103,18 @@ markBoringParentAccounts = tieAccountParents . mapAccounts mark
            | otherwise = a
 
 balanceReportItem :: ReportOpts -> Query -> Account -> BalanceReportItem
-balanceReportItem opts _ a@Account{aname=name}
+balanceReportItem opts q a
   | flat_ opts = ((name, name,       0),      (if flatShowsExclusiveBalance then aebalance else aibalance) a)
   | otherwise  = ((name, elidedname, indent), aibalance a)
   where
+    name | queryDepth q > 0 = aname a
+         | otherwise        = "..."
     elidedname = accountNameFromComponents (adjacentboringparentnames ++ [accountLeafName name])
     adjacentboringparentnames = reverse $ map (accountLeafName.aname) $ takeWhile aboring $ parents
     indent = length $ filter (not.aboring) parents
-    parents = init $ parentAccounts a
+    -- parents exclude the tree's root node
+    parents = case parentAccounts a of [] -> []
+                                       as -> init as
 
 -- -- the above using the newer multi balance report code:
 -- balanceReport' opts q j = (items, total)
@@ -116,13 +124,14 @@ balanceReportItem opts _ a@Account{aname=name}
 --     total = headDef 0 mbrtotals
 
 tests_balanceReport =
-  let (opts,journal) `gives` r = do
-         let (eitems, etotal) = r
-             (aitems, atotal) = balanceReport opts (queryFromOpts nulldate opts) journal
-         assertEqual "items" eitems aitems
-         -- assertEqual "" (length eitems) (length aitems)
-         -- mapM (\(e,a) -> assertEqual "" e a) $ zip eitems aitems
-         assertEqual "total" etotal atotal
+  let
+    (opts,journal) `gives` r = do
+      let (eitems, etotal) = r
+          (aitems, atotal) = balanceReport opts (queryFromOpts nulldate opts) journal
+          showw (acct,amt) = (acct, showMixedAmountDebug amt)
+      assertEqual "items" (map showw eitems) (map showw aitems)
+      assertEqual "total" (showMixedAmountDebug etotal) (showMixedAmountDebug atotal)
+    usd0 = nullamt{acommodity="$"}
   in [
 
    "balanceReport with no args on null journal" ~: do
@@ -142,7 +151,7 @@ tests_balanceReport =
      ,(("income:salary","salary",1), mamountp' "$-1.00")
      ,(("liabilities:debts","liabilities:debts",0), mamountp' "$1.00")
      ],
-     Mixed [nullamt])
+     Mixed [usd0])
 
   ,"balanceReport with --depth=N" ~: do
    (defreportopts{depth_=Just 1}, samplejournal) `gives`
@@ -152,7 +161,7 @@ tests_balanceReport =
      ,(("income",      "income",      0), mamountp' "$-2.00")
      ,(("liabilities", "liabilities", 0), mamountp'  "$1.00")
      ],
-     Mixed [nullamt])
+     Mixed [usd0])
 
   ,"balanceReport with depth:N" ~: do
    (defreportopts{query_="depth:1"}, samplejournal) `gives`
@@ -162,18 +171,18 @@ tests_balanceReport =
      ,(("income",      "income",      0), mamountp' "$-2.00")
      ,(("liabilities", "liabilities", 0), mamountp'  "$1.00")
      ],
-     Mixed [nullamt])
+     Mixed [usd0])
 
   ,"balanceReport with a date or secondary date span" ~: do
    (defreportopts{query_="date:'in 2009'"}, samplejournal2) `gives`
     ([],
      Mixed [nullamt])
-   (defreportopts{query_="edate:'in 2009'"}, samplejournal2) `gives`
+   (defreportopts{query_="date2:'in 2009'"}, samplejournal2) `gives`
     ([
       (("assets:bank:checking","assets:bank:checking",0),mamountp' "$1.00")
      ,(("income:salary","income:salary",0),mamountp' "$-1.00")
      ],
-     Mixed [nullamt])
+     Mixed [usd0])
 
   ,"balanceReport with desc:" ~: do
    (defreportopts{query_="desc:income"}, samplejournal) `gives`
@@ -181,13 +190,13 @@ tests_balanceReport =
       (("assets:bank:checking","assets:bank:checking",0),mamountp' "$1.00")
      ,(("income:salary","income:salary",0), mamountp' "$-1.00")
      ],
-     Mixed [nullamt])
+     Mixed [usd0])
 
   ,"balanceReport with not:desc:" ~: do
    (defreportopts{query_="not:desc:income"}, samplejournal) `gives`
     ([
       (("assets","assets",0), mamountp' "$-2.00")
-     ,(("assets:bank","bank",1), Mixed [nullamt])
+     ,(("assets:bank","bank",1), Mixed [usd0])
      ,(("assets:bank:checking","checking",2),mamountp' "$-1.00")
      ,(("assets:bank:saving","saving",2), mamountp' "$1.00")
      ,(("assets:cash","cash",1), mamountp' "$-2.00")
@@ -197,7 +206,7 @@ tests_balanceReport =
      ,(("income:gifts","income:gifts",0), mamountp' "$-1.00")
      ,(("liabilities:debts","liabilities:debts",0), mamountp' "$1.00")
      ],
-     Mixed [nullamt])
+     Mixed [usd0])
 
 
 {-
@@ -246,7 +255,7 @@ tests_balanceReport =
      ,"                   0"
      ]
 
-    ,"accounts report with unmatched parent of two matched subaccounts" ~: 
+    ,"accounts report with unmatched parent of two matched subaccounts" ~:
      defreportopts{patterns_=["cash","saving"]} `gives`
      ["                 $-1  assets"
      ,"                  $1    bank:saving"
@@ -255,7 +264,7 @@ tests_balanceReport =
      ,"                 $-1"
      ]
 
-    ,"accounts report with multi-part account name" ~: 
+    ,"accounts report with multi-part account name" ~:
      defreportopts{patterns_=["expenses:food"]} `gives`
      ["                  $1  expenses:food"
      ,"--------------------"
@@ -275,13 +284,13 @@ tests_balanceReport =
      ,"                  $1"
      ]
 
-    ,"accounts report negative account pattern always matches full name" ~: 
+    ,"accounts report negative account pattern always matches full name" ~:
      defreportopts{patterns_=["not:e"]} `gives`
      ["--------------------"
      ,"                   0"
      ]
 
-    ,"accounts report negative patterns affect totals" ~: 
+    ,"accounts report negative patterns affect totals" ~:
      defreportopts{patterns_=["expenses","not:food"]} `gives`
      ["                  $1  expenses:supplies"
      ,"--------------------"
@@ -316,10 +325,11 @@ tests_balanceReport =
 -}
  ]
 
-Right samplejournal2 = journalBalanceTransactions $ 
+Right samplejournal2 = journalBalanceTransactions $
          nulljournal
          {jtxns = [
            txnTieKnot $ Transaction {
+             tsourcepos=nullsourcepos,
              tdate=parsedate "2008/01/01",
              tdate2=Just $ parsedate "2009/01/01",
              tstatus=False,
@@ -335,12 +345,12 @@ Right samplejournal2 = journalBalanceTransactions $
            }
           ]
          }
-         
+
 -- tests_isInterestingIndented = [
---   "isInterestingIndented" ~: do 
+--   "isInterestingIndented" ~: do
 --    let (opts, journal, acctname) `gives` r = isInterestingIndented opts l acctname `is` r
 --           where l = ledgerFromJournal (queryFromOpts nulldate opts) journal
-     
+
 --    (defreportopts, samplejournal, "expenses") `gives` True
 --  ]
 
