@@ -1,4 +1,4 @@
-{-# LANGUAGE RecordWildCards, DeriveDataTypeable, FlexibleInstances #-}
+{-# LANGUAGE RecordWildCards, DeriveDataTypeable, FlexibleInstances, TupleSections #-}
 {-|
 
 Postings report, used by the register command.
@@ -35,10 +35,17 @@ import Hledger.Reports.ReportOptions
 type PostingsReport = (String               -- label for the running balance column XXX remove
                       ,[PostingsReportItem] -- line items, one per posting
                       )
-type PostingsReportItem = (Maybe Day    -- posting date, if this is the first posting in a transaction or if it's different from the previous posting's date
-                          ,Maybe String -- transaction description, if this is the first posting in a transaction
-                          ,Posting      -- the posting, possibly with account name depth-clipped
-                          ,MixedAmount  -- the running total after this posting (or with --average, the running average)
+type PostingsReportItem = (Maybe Day    -- The posting date, if this is the first posting in a
+                                        -- transaction or if it's different from the previous
+                                        -- posting's date. Or if this a summary posting, the
+                                        -- report interval's start date if this is the first
+                                        -- summary posting in the interval.
+                          ,Maybe Day    -- If this is a summary posting, the report interval's
+                                        -- end date if this is the first summary posting in
+                                        -- the interval.
+                          ,Maybe String -- The posting's transaction's description, if this is the first posting in the transaction.
+                          ,Posting      -- The posting, possibly with the account name depth-clipped.
+                          ,MixedAmount  -- The running total after this posting (or with --average, the running average).
                           )
 
 -- | Select postings from the journal and add running balance and other
@@ -50,11 +57,13 @@ postingsReport opts q j = (totallabel, items)
       symq = dbg "symq"   $ filterQuery queryIsSym $ dbg "requested q" q
       depth = queryDepth q
       depthless = filterQuery (not . queryIsDepth)
-      datelessq = filterQuery (not . queryIsDate) q
-      (dateqcons,pdate) | date2_ opts = (Date2, postingDate2)
-                        | otherwise   = (Date, postingDate)
-      requestedspan  = dbg "requestedspan"  $ queryDateSpan (date2_ opts) q                              -- span specified by -b/-e/-p options and query args
-      requestedspan' = dbg "requestedspan'" $ requestedspan `spanDefaultsFrom` journalDateSpan (date2_ opts) j  -- if open-ended, close it using the journal's end dates
+      datelessq = filterQuery (not . queryIsDateOrDate2) q
+      -- XXX date:/date2:/--date2 handling is not robust, combinations of these can confuse it
+      dateq = filterQuery queryIsDateOrDate2 q
+      (dateqcons,pdate) | queryIsDate2 dateq || (queryIsDate dateq && date2_ opts) = (Date2, postingDate2)
+                        | otherwise = (Date, postingDate)
+      requestedspan  = dbg "requestedspan"  $ queryDateSpan' q   -- span specified by -b/-e/-p options and query args
+      requestedspan' = dbg "requestedspan'" $ requestedspan `spanDefaultsFrom` journalDateSpan ({-date2_ opts-} False) j  -- if open-ended, close it using the journal's end dates
       intervalspans  = dbg "intervalspans"  $ splitSpan (intervalFromOpts opts) requestedspan' -- interval spans enclosing it
       reportstart    = dbg "reportstart"    $ maybe Nothing spanStart $ headMay intervalspans
       reportend      = dbg "reportend"      $ maybe Nothing spanEnd   $ lastMay intervalspans
@@ -63,7 +72,7 @@ postingsReport opts q j = (totallabel, items)
       beforeendq     = dbg "beforeendq"     $ dateqcons $ DateSpan Nothing reportend
       reportq        = dbg "reportq"        $ depthless $ And [datelessq, beforeendq] -- user's query with no start date, end date on an interval boundary and no depth limit
 
-      pstoend = 
+      pstoend =
           dbg "ps4" $ sortBy (comparing pdate) $                                  -- sort postings by date (or date2)
           dbg "ps3" $ map (filterPostingAmount symq) $                            -- remove amount parts which the query's cur: terms would exclude
           dbg "ps2" $ (if related_ opts then concatMap relatedPostings else id) $ -- with -r, replace each with its sibling postings
@@ -71,14 +80,14 @@ postingsReport opts q j = (totallabel, items)
                       journalPostings $ journalSelectingAmountFromOpts opts j
       (precedingps, reportps) = dbg "precedingps, reportps" $ span (beforestartq `matchesPosting`) pstoend
 
-      empty = queryEmpty q
+      showempty = queryEmpty q || average_ opts
       -- displayexpr = display_ opts  -- XXX
       interval = intervalFromOpts opts -- XXX
 
       whichdate = whichDateFromOpts opts
-      itemps | interval == NoInterval = reportps
-             | otherwise              = summarisePostingsByInterval interval whichdate depth empty reportspan reportps
-      items = postingsReportItems itemps nullposting whichdate depth startbal runningcalc 1
+      itemps | interval == NoInterval = map (,Nothing) reportps
+             | otherwise              = summarisePostingsByInterval interval whichdate depth showempty reportspan reportps
+      items = dbg "items" $ postingsReportItems itemps (nullposting,Nothing) whichdate depth startbal runningcalc 1
         where
           startbal = if balancetype_ opts == HistoricalBalance then sumPostings precedingps else 0
           runningcalc | average_ opts = \i avg amt -> avg + (amt - avg) `divideMixedAmount` (fromIntegral i) -- running average
@@ -89,37 +98,43 @@ postingsReport opts q j = (totallabel, items)
 
 totallabel = "Total"
 
--- | Generate postings report line items.
-postingsReportItems :: [Posting] -> Posting -> WhichDate -> Int -> MixedAmount -> (Int -> MixedAmount -> MixedAmount -> MixedAmount) -> Int -> [PostingsReportItem]
+-- | Generate postings report line items from a list of postings or (with
+-- non-Nothing dates attached) summary postings.
+postingsReportItems :: [(Posting,Maybe Day)] -> (Posting,Maybe Day) -> WhichDate -> Int -> MixedAmount -> (Int -> MixedAmount -> MixedAmount -> MixedAmount) -> Int -> [PostingsReportItem]
 postingsReportItems [] _ _ _ _ _ _ = []
-postingsReportItems (p:ps) pprev wd d b runningcalcfn itemnum = i:(postingsReportItems ps p wd d b' runningcalcfn (itemnum+1))
+postingsReportItems ((p,menddate):ps) (pprev,menddateprev) wd d b runningcalcfn itemnum = i:(postingsReportItems ps (p,menddate) wd d b' runningcalcfn (itemnum+1))
     where
-      i = mkpostingsReportItem showdate showdesc wd p' b'
-      showdate = isfirstintxn || isdifferentdate
-      showdesc = isfirstintxn
+      i = mkpostingsReportItem showdate showdesc wd menddate p' b'
+      (showdate, showdesc) | isJust menddate = (menddate /= menddateprev,        False)
+                           | otherwise       = (isfirstintxn || isdifferentdate, isfirstintxn)
       isfirstintxn = ptransaction p /= ptransaction pprev
       isdifferentdate = case wd of PrimaryDate   -> postingDate p  /= postingDate pprev
                                    SecondaryDate -> postingDate2 p /= postingDate2 pprev
-      p' = p{paccount=clipAccountName d $ paccount p}
+      p' = p{paccount= clipOrEllipsifyAccountName d $ paccount p}
       b' = runningcalcfn itemnum b (pamount p)
 
 -- | Generate one postings report line item, containing the posting,
 -- the current running balance, and optionally the posting date and/or
 -- the transaction description.
-mkpostingsReportItem :: Bool -> Bool -> WhichDate -> Posting -> MixedAmount -> PostingsReportItem
-mkpostingsReportItem showdate showdesc wd p b = (if showdate then Just date else Nothing, if showdesc then Just desc else Nothing, p, b)
-    where
-      date = case wd of PrimaryDate   -> postingDate p
-                        SecondaryDate -> postingDate2 p
-      desc = maybe "" tdescription $ ptransaction p
+mkpostingsReportItem :: Bool -> Bool -> WhichDate -> Maybe Day -> Posting -> MixedAmount -> PostingsReportItem
+mkpostingsReportItem showdate showdesc wd menddate p b =
+  (if showdate then Just date else Nothing
+  ,menddate
+  ,if showdesc then Just desc else Nothing
+  ,p
+  ,b
+  )
+  where
+    date = case wd of PrimaryDate   -> postingDate p
+                      SecondaryDate -> postingDate2 p
+    desc = maybe "" tdescription $ ptransaction p
 
--- | Convert a list of postings into summary postings. Summary postings
--- are one per account per interval and aggregated to the specified depth
--- if any.
-summarisePostingsByInterval :: Interval -> WhichDate -> Int -> Bool -> DateSpan -> [Posting] -> [Posting]
-summarisePostingsByInterval interval wd depth empty reportspan ps = concatMap summarisespan $ splitSpan interval reportspan
+-- | Convert a list of postings into summary postings, one per interval,
+-- aggregated to the specified depth if any.
+summarisePostingsByInterval :: Interval -> WhichDate -> Int -> Bool -> DateSpan -> [Posting] -> [SummaryPosting]
+summarisePostingsByInterval interval wd depth showempty reportspan ps = concatMap summarisespan $ splitSpan interval reportspan
     where
-      summarisespan s = summarisePostingsInDateSpan s wd depth empty (postingsinspan s)
+      summarisespan s = summarisePostingsInDateSpan s wd depth showempty (postingsinspan s)
       postingsinspan s = filter (isPostingInDateSpan' wd s) ps
 
 tests_summarisePostingsByInterval = [
@@ -127,37 +142,43 @@ tests_summarisePostingsByInterval = [
     summarisePostingsByInterval (Quarters 1) PrimaryDate 99999 False (DateSpan Nothing Nothing) [] ~?= []
  ]
 
+-- | A summary posting summarises the activity in one account within a report
+-- interval. It is currently kludgily represented by a regular Posting with no
+-- description, the interval's start date stored as the posting date, and the
+-- interval's end date attached with a tuple.
+type SummaryPosting = (Posting, Maybe Day)
+
 -- | Given a date span (representing a reporting interval) and a list of
--- postings within it: aggregate the postings so there is only one per
--- account, and adjust their date/description so that they will render
--- as a summary for this interval.
---
--- As usual with date spans the end date is exclusive, but for display
--- purposes we show the previous day as end date, like ledger.
+-- postings within it, aggregate the postings into one summary posting per
+-- account.
 --
 -- When a depth argument is present, postings to accounts of greater
--- depth are aggregated where possible.
+-- depth are also aggregated where possible. If the depth is 0, all
+-- postings in the span are aggregated into a single posting with
+-- account name "...".
 --
 -- The showempty flag includes spans with no postings and also postings
 -- with 0 amount.
-summarisePostingsInDateSpan :: DateSpan -> WhichDate -> Int -> Bool -> [Posting] -> [Posting]
+--
+summarisePostingsInDateSpan :: DateSpan -> WhichDate -> Int -> Bool -> [Posting] -> [SummaryPosting]
 summarisePostingsInDateSpan (DateSpan b e) wd depth showempty ps
     | null ps && (isNothing b || isNothing e) = []
-    | null ps && showempty = [summaryp]
-    | otherwise = summaryps'
+    | null ps && showempty = [(summaryp, Just e')]
+    | otherwise = summarypes
     where
-      summaryp = summaryPosting b' ("- "++ showDate (addDays (-1) e'))
+      postingdate = if wd == PrimaryDate then postingDate else postingDate2
       b' = fromMaybe (maybe nulldate postingdate $ headMay ps) b
       e' = fromMaybe (maybe (addDays 1 nulldate) postingdate $ lastMay ps) e
-      postingdate = if wd == PrimaryDate then postingDate else postingDate2
-      summaryPosting date desc = nullposting{ptransaction=Just nulltransaction{tdate=date,tdescription=desc}}
-      summaryps' = (if showempty then id else filter (not . isZeroMixedAmount . pamount)) summaryps
-      summaryps = [summaryp{paccount=a,pamount=balance a} | a <- clippedanames]
-      clippedanames = nub $ map (clipAccountName depth) anames
+      summaryp = nullposting{pdate=Just b'}
+      clippedanames | depth > 0 = nub $ map (clipAccountName depth) anames
+                    | otherwise = ["..."]
+      summaryps | depth > 0 = [summaryp{paccount=a,pamount=balance a} | a <- clippedanames]
+                | otherwise = [summaryp{paccount="...",pamount=sum $ map pamount ps}]
+      summarypes = map (, Just e') $ (if showempty then id else filter (not . isZeroMixedAmount . pamount)) summaryps
       anames = sort $ nub $ map paccount ps
       -- aggregate balances by account, like ledgerFromJournal, then do depth-clipping
       accts = accountsFromPostings ps
-      balance a = maybe nullmixedamt bal $ lookupAccount a accts 
+      balance a = maybe nullmixedamt bal $ lookupAccount a accts
         where
           bal = if isclipped a then aibalance else aebalance
           isclipped a = accountNameLevel a >= depth
@@ -247,7 +268,7 @@ tests_postingsReport = [
      ]
 
   ,"postings report with cleared option" ~:
-   do 
+   do
     let opts = defreportopts{cleared_=True}
     j <- readJournal' sample_journal_str
     (postingsReportAsText opts $ postingsReport opts (queryFromOpts date1 opts) j) `is` unlines
@@ -259,7 +280,7 @@ tests_postingsReport = [
      ]
 
   ,"postings report with uncleared option" ~:
-   do 
+   do
     let opts = defreportopts{uncleared_=True}
     j <- readJournal' sample_journal_str
     (postingsReportAsText opts $ postingsReport opts (queryFromOpts date1 opts) j) `is` unlines
@@ -272,7 +293,7 @@ tests_postingsReport = [
      ]
 
   ,"postings report sorts by date" ~:
-   do 
+   do
     j <- readJournal' $ unlines
         ["2008/02/02 a"
         ,"  b  1"
@@ -294,7 +315,7 @@ tests_postingsReport = [
      ]
 
   ,"postings report with account pattern, case insensitive" ~:
-   do 
+   do
     j <- samplejournal
     let opts = defreportopts{patterns_=["cAsH"]}
     (postingsReportAsText opts $ postingsReport opts (queryFromOpts date1 opts) j) `is` unlines
@@ -302,9 +323,9 @@ tests_postingsReport = [
      ]
 
   ,"postings report with display expression" ~:
-   do 
+   do
     j <- samplejournal
-    let gives displayexpr = 
+    let gives displayexpr =
             (registerdates (postingsReportAsText opts $ postingsReport opts (queryFromOpts date1 opts) j) `is`)
                 where opts = defreportopts{display_=Just displayexpr}
     "d<[2008/6/2]"  `gives` ["2008/01/01","2008/06/01"]
@@ -314,7 +335,7 @@ tests_postingsReport = [
     "d>[2008/6/2]"  `gives` ["2008/06/03","2008/12/31"]
 
   ,"postings report with period expression" ~:
-   do 
+   do
     j <- samplejournal
     let periodexpr `gives` dates = do
           j' <- samplejournal
@@ -344,7 +365,7 @@ tests_postingsReport = [
   ]
 
   , "postings report with depth arg" ~:
-   do 
+   do
     j <- samplejournal
     let opts = defreportopts{depth_=Just 2}
     (postingsReportAsText opts $ postingsReport opts (queryFromOpts date1 opts) j) `is` unlines
