@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE FlexibleContexts #-}
 {-|
 
 Standard imports and utilities which are useful everywhere, or needed low
@@ -14,69 +14,62 @@ module Hledger.Utils (---- provide these frequently used modules - or not, for c
                           -- module Data.Time.Clock,
                           -- module Data.Time.LocalTime,
                           -- module Data.Tree,
-                          -- module Debug.Trace,
                           -- module Text.RegexPR,
                           -- module Test.HUnit,
                           -- module Text.Printf,
                           ---- all of this one:
                           module Hledger.Utils,
-                          Debug.Trace.trace,
+                          module Hledger.Utils.Debug,
+                          module Hledger.Utils.Regex,
+                          -- Debug.Trace.trace,
                           -- module Data.PPrint,
                           -- module Hledger.Utils.UTF8IOCompat
                           SystemString,fromSystemString,toSystemString,error',userError',
-#if __GLASGOW_HASKELL__ >= 704
-                          ppShow
-#endif
                           -- the rest need to be done in each module I think
                           )
 where
-import Control.Monad (liftM, when)
+import Control.Monad (liftM)
 import Control.Monad.Error (MonadIO)
 import Control.Monad.IO.Class (liftIO)
 import Data.Char
 import Data.List
 import qualified Data.Map as M
-import Data.Maybe
+-- import Data.Maybe
 -- import Data.PPrint
 import Data.Time.Clock
 import Data.Time.LocalTime
 import Data.Tree
-import Debug.Trace
-import Safe (readDef)
 import System.Directory (getHomeDirectory)
-import System.Environment (getArgs)
-import System.Exit
 import System.FilePath((</>), isRelative)
 import System.IO
-import System.IO.Unsafe (unsafePerformIO)
 import Test.HUnit
-import Text.ParserCombinators.Parsec
+import Text.Parsec
 import Text.Printf
-import Text.Regex.TDFA
-import Text.RegexPR
 -- import qualified Data.Map as Map
--- 
+
+import Hledger.Utils.Debug
+import Hledger.Utils.Regex
 -- import Prelude hiding (readFile,writeFile,appendFile,getContents,putStr,putStrLn)
 -- import Hledger.Utils.UTF8IOCompat   (readFile,writeFile,appendFile,getContents,putStr,putStrLn)
 import Hledger.Utils.UTF8IOCompat (SystemString,fromSystemString,toSystemString,error',userError')
 
-#if __GLASGOW_HASKELL__ >= 704
-import Text.Show.Pretty (ppShow)
-#else
--- the required pretty-show version requires GHC >= 7.4
-ppShow :: Show a => a -> String
-ppShow = show
-#endif
-
 -- strings
 
 lowercase = map toLower
 uppercase = map toUpper
 
+-- | Remove leading and trailing whitespace.
 strip = lstrip . rstrip
-lstrip = dropWhile (`elem` " \t") :: String -> String
+
+-- | Remove leading whitespace.
+lstrip = dropWhile (`elem` " \t") :: String -> String -- XXX isSpace ?
+
+-- | Remove trailing whitespace.
 rstrip = reverse . lstrip . reverse
 
+-- | Remove trailing newlines/carriage returns.
+chomp = reverse . dropWhile (`elem` "\r\n") . reverse
+
 stripbrackets = dropWhile (`elem` "([") . reverse . dropWhile (`elem` "])") . reverse :: String -> String
 
 elideLeft width s =
@@ -248,54 +241,17 @@ fifth5  (_,_,_,_,x) = x
 difforzero :: (Num a, Ord a) => a -> a -> a
 difforzero a b = maximum [(a - b), 0]
 
--- regexps
--- Note many of these will die on malformed regexps.
-
--- regexMatch :: String -> String -> MatchFun Maybe
-regexMatch r s = matchRegexPR r s
-
--- regexMatchCI :: String -> String -> MatchFun Maybe
-regexMatchCI r s = regexMatch (regexToCaseInsensitive r) s
-
-regexMatches :: String -> String -> Bool
-regexMatches r s = isJust $ matchRegexPR r s
-
-regexMatchesCI :: String -> String -> Bool
-regexMatchesCI r s = regexMatches (regexToCaseInsensitive r) s
-
-containsRegex = regexMatchesCI
-
-regexReplace :: String -> String -> String -> String
-regexReplace r repl s = gsubRegexPR r repl s
-
-regexReplaceCI :: String -> String -> String -> String
-regexReplaceCI r s = regexReplace (regexToCaseInsensitive r) s
-
-regexReplaceBy :: String -> (String -> String) -> String -> String
-regexReplaceBy r replfn s = gsubRegexPRBy r replfn s
-
-regexToCaseInsensitive :: String -> String
-regexToCaseInsensitive r = "(?i)"++ r
-
-regexSplit :: String -> String -> [String]
-regexSplit = splitRegexPR
-
--- regex-compat (regex-posix) functions that perform better than regexpr.
-regexMatchesRegexCompat :: String -> String -> Bool
-regexMatchesRegexCompat = flip (=~)
-
-regexMatchesCIRegexCompat :: String -> String -> Bool
-regexMatchesCIRegexCompat r = match (makeRegexOpts defaultCompOpt { multiline = True, caseSensitive = False, newSyntax = True } defaultExecOpt r)
-
 -- lists
 
 splitAtElement :: Eq a => a -> [a] -> [[a]]
-splitAtElement e l = 
-    case dropWhile (e==) l of
-      [] -> []
-      l' -> first : splitAtElement e rest
-        where
-          (first,rest) = break (e==) l'
+splitAtElement x l =
+  case l of
+    []          -> []
+    e:es | e==x -> split es
+    es          -> split es
+  where
+    split es = let (first,rest) = break (x==) es
+               in first : splitAtElement x rest
 
 -- trees
 
@@ -324,7 +280,7 @@ subtreeinforest _ [] = Nothing
 subtreeinforest v (t:ts) = case (subtreeat v t) of
                              Just t' -> Just t'
                              Nothing -> subtreeinforest v ts
-          
+
 -- | remove all nodes past a certain depth
 treeprune :: Int -> Tree a -> Tree a
 treeprune 0 t = Node (root t) []
@@ -336,15 +292,15 @@ treemap f t = Node (f $ root t) (map (treemap f) $ branches t)
 
 -- | remove all subtrees whose nodes do not fulfill predicate
 treefilter :: (a -> Bool) -> Tree a -> Tree a
-treefilter f t = Node 
-                 (root t) 
+treefilter f t = Node
+                 (root t)
                  (map (treefilter f) $ filter (treeany f) $ branches t)
-    
+
 -- | is predicate true in any node of tree ?
 treeany :: (a -> Bool) -> Tree a -> Bool
 treeany f t = f (root t) || any (treeany f) (branches t)
-    
--- treedrop -- remove the leaves which do fulfill predicate. 
+
+-- treedrop -- remove the leaves which do fulfill predicate.
 -- treedropall -- do this repeatedly.
 
 -- | show a compact ascii representation of a tree
@@ -374,146 +330,18 @@ treeFromPaths :: (Ord a) => [[a]] -> FastTree a
 treeFromPaths = foldl' mergeTrees emptyTree . map treeFromPath
 
 
--- debugging
-
--- more:
--- http://hackage.haskell.org/packages/archive/TraceUtils/0.1.0.2/doc/html/Debug-TraceUtils.html
--- http://hackage.haskell.org/packages/archive/trace-call/0.1/doc/html/Debug-TraceCall.html
--- http://hackage.haskell.org/packages/archive/htrace/0.1/doc/html/Debug-HTrace.html
--- http://hackage.haskell.org/packages/archive/traced/2009.7.20/doc/html/Debug-Traced.html
-
--- | Trace (print on stdout at runtime) a showable value.
--- (for easily tracing in the middle of a complex expression)
-strace :: Show a => a -> a
-strace a = trace (show a) a
-
--- | Labelled trace - like strace, with a label prepended.
-ltrace :: Show a => String -> a -> a
-ltrace l a = trace (l ++ ": " ++ show a) a
-
--- | Monadic trace - like strace, but works as a standalone line in a monad.
-mtrace :: (Monad m, Show a) => a -> m a
-mtrace a = strace a `seq` return a
-
--- | Custom trace - like strace, with a custom show function.
-traceWith :: (a -> String) -> a -> a
-traceWith f e = trace (f e) e
-
--- | Parsec trace - show the current parsec position and next input,
--- and the provided label if it's non-null.
-ptrace :: String -> GenParser Char st ()
-ptrace msg = do
-  pos <- getPosition
-  next <- take peeklength `fmap` getInput
-  let (l,c) = (sourceLine pos, sourceColumn pos)
-      s  = printf "at line %2d col %2d: %s" l c (show next) :: String
-      s' = printf ("%-"++show (peeklength+30)++"s") s ++ " " ++ msg
-  trace s' $ return ()
-  where
-    peeklength = 30
-
--- | Global debug level, which controls the verbosity of debug output
--- on the console. The default is 0 meaning no debug output. The
--- @--debug@ command line flag sets it to 1, or @--debug=N@ sets it to
--- a higher value (note: not @--debug N@ for some reason).  This uses
--- unsafePerformIO and can be accessed from anywhere and before normal
--- command-line processing. After command-line processing, it is also
--- available as the @debug_@ field of 'Hledger.Cli.Options.CliOpts'.
-debugLevel :: Int
-debugLevel = case snd $ break (=="--debug") args of
-               "--debug":[]  -> 1
-               "--debug":n:_ -> readDef 1 n
-               _             ->
-                 case take 1 $ filter ("--debug" `isPrefixOf`) args of
-                   ['-':'-':'d':'e':'b':'u':'g':'=':v] -> readDef 1 v
-                   _                                   -> 0
-
-    where
-      args = unsafePerformIO getArgs
-
--- | Print a message and a showable value to the console if the global
--- debug level is non-zero.  Uses unsafePerformIO.
-dbg :: Show a => String -> a -> a
-dbg = dbg1
-
-dbg0 :: Show a => String -> a -> a
-dbg0 = dbgAt 0
-
-dbg1 :: Show a => String -> a -> a
-dbg1 = dbgAt 1
-
-dbg2 :: Show a => String -> a -> a
-dbg2 = dbgAt 2
-
--- | Print a message and a showable value to the console if the global
--- debug level is at or above the specified level.  Uses unsafePerformIO.
-dbgAt :: Show a => Int -> String -> a -> a
-dbgAt lvl = dbgppshow lvl
-
-dbgAtM :: Show a => Int -> String -> a -> IO ()
-dbgAtM lvl lbl x = dbgAt lvl lbl x `seq` return ()
-
--- | Print a showable value to the console, with a message, if the
--- debug level is at or above the specified level (uses
--- unsafePerformIO).
--- Values are displayed with show, all on one line, which is hard to read.
-dbgshow :: Show a => Int -> String -> a -> a
-dbgshow level
-    | debugLevel >= level = ltrace
-    | otherwise           = flip const
-
--- | Print a showable value to the console, with a message, if the
--- debug level is at or above the specified level (uses
--- unsafePerformIO).
--- Values are displayed with ppShow, each field/constructor on its own line.
-dbgppshow :: Show a => Int -> String -> a -> a
-dbgppshow level
-    | debugLevel < level = flip const
-    | otherwise = \s a -> let p = ppShow a
-                              ls = lines p
-                              nlorspace | length ls > 1 = "\n"
-                                        | otherwise     = " " ++ take (10 - length s) (repeat ' ')
-                              ls' | length ls > 1 = map (" "++) ls
-                                  | otherwise     = ls
-                          in trace (s++":"++nlorspace++intercalate "\n" ls') a
-
--- -- | Print a showable value to the console, with a message, if the
--- -- debug level is at or above the specified level (uses
--- -- unsafePerformIO).
--- -- Values are displayed with pprint. Field names are not shown, but the
--- -- output is compact with smart line wrapping, long data elided,
--- -- and slow calculations timed out.
--- dbgpprint :: Data a => Int -> String -> a -> a
--- dbgpprint level msg a
---     | debugLevel >= level = unsafePerformIO $ do
---                               pprint a >>= putStrLn . ((msg++": \n") ++) . show
---                               return a
---     | otherwise           = a
-
-
--- | Like dbg, then exit the program. Uses unsafePerformIO.
-dbgExit :: Show a => String -> a -> a
-dbgExit msg = const (unsafePerformIO exitFailure) . dbg msg
-
--- | Print a message and parsec debug info (parse position and next
--- input) to the console when the debug level is at or above
--- this level. Uses unsafePerformIO.
--- pdbgAt :: GenParser m => Float -> String -> m ()
-pdbg level msg = when (level <= debugLevel) $ ptrace msg
-
-
 -- parsing
 
 -- | Backtracking choice, use this when alternatives share a prefix.
 -- Consumes no input if all choices fail.
-choice' :: [GenParser tok st a] -> GenParser tok st a
-choice' = choice . map Text.ParserCombinators.Parsec.try
+choice' :: Stream s m t => [ParsecT s u m a] -> ParsecT s u m a
+choice' = choice . map Text.Parsec.try
 
-parsewith :: Parser a -> String -> Either ParseError a
-parsewith p = parse p ""
+parsewith :: Parsec [Char] () a -> String -> Either ParseError a
+parsewith p = runParser p () ""
 
-parseWithCtx :: b -> GenParser Char b a -> String -> Either ParseError a
-parseWithCtx ctx p = runParser p ctx ""
+parseWithCtx :: Stream s m t => u -> ParsecT s u m a -> s -> m (Either ParseError a)
+parseWithCtx ctx p = runParserT p ctx ""
 
 fromparse :: Either ParseError a -> a
 fromparse = either parseerror id
@@ -527,16 +355,16 @@ showParseError e = "parse error at " ++ show e
 showDateParseError :: ParseError -> String
 showDateParseError e = printf "date parse error (%s)" (intercalate ", " $ tail $ lines $ show e)
 
-nonspace :: GenParser Char st Char
+nonspace :: (Stream [Char] m Char) => ParsecT [Char] st m Char
 nonspace = satisfy (not . isSpace)
 
-spacenonewline :: GenParser Char st Char
+spacenonewline :: (Stream [Char] m Char) => ParsecT [Char] st m Char
 spacenonewline = satisfy (`elem` " \v\f\t")
 
-restofline :: GenParser Char st String
+restofline :: (Stream [Char] m Char) => ParsecT [Char] st m String
 restofline = anyChar `manyTill` newline
 
-eolof :: GenParser Char st ()
+eolof :: (Stream [Char] m Char) => ParsecT [Char] st m ()
 eolof = (newline >> return ()) <|> eof
 
 -- time
@@ -599,7 +427,7 @@ applyN :: Int -> (a -> a) -> a -> a
 applyN n f = (!! n) . iterate f
 
 -- | Convert a possibly relative, possibly tilde-containing file path to an absolute one,
--- given the current directory. ~username is not supported. Leave "-" unchanged. 
+-- given the current directory. ~username is not supported. Leave "-" unchanged.
 expandPath :: MonadIO m => FilePath -> FilePath -> m FilePath -- general type sig for use in reader parsers
 expandPath _ "-" = return "-"
 expandPath curdir p = (if isRelative p then (curdir </>) else id) `liftM` expandPath' p
